import copy
from typing import Any
from typing import Dict
from typing import List

STATE_NAME = "aws.neptune.db_cluster_parameter_group"


async def search_raw(
    hub, ctx, name: str, resource_id: str = None, filters: List = None
):
    """
    Fetch one or more neptune db_cluster_parameter_group objects in its raw form.

    :param hub: idem hub
    :param ctx: idem context
    :param name: (string) idem name
    :param resource_id: (string) resource_id maps to DBClusterParameterGroupName attribute for AWS neptune service
    :param filters: This is not supported for AWS neptune yet as per AWS documentation.
    :return: Raw representation of one or more AWS neptune DBClusterParameterGroup objects.
    """
    result = dict(comment=[], ret=None, result=True)
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["result"] = False
        result["comment"] = list(syntax_validation["comment"])
        return result
    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )
    ret = await hub.exec.boto3.client.neptune.describe_db_cluster_parameter_groups(
        ctx,
        DBClusterParameterGroupName=resource_id if resource_id else None,
        Filters=boto3_filter,
    )
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result


async def update(
    hub,
    ctx,
    name: str,
    resource_id: str,
    desired_state: Dict,
    is_created: bool,
) -> Dict[str, Any]:
    """
    Updates the db_cluster_parameter_group represented by resource_id which is the db_cluster_parameter_group_name to the desired state.
    If is_created flag is set to True, the comment in result is not updated as it is assumed that
    the resource is being updated right after it was created to apply remaining parameters.
    Only following parameters can be modified:
    - parameters
    - tags
    """
    result = dict(comment=(), old_state=None, new_state=None, result=True)
    if ctx.get("test", False) and is_created:
        current_state = desired_state
    else:
        current_state_result = (
            await hub.exec.aws.neptune.db_cluster_parameter_group.get(
                ctx=ctx, name=name, resource_id=resource_id
            )
        )
        if not (
            current_state_result
            and current_state_result["result"]
            and current_state_result["ret"]
        ):
            result["result"] = False
            result["comment"] = current_state_result["comment"]
            return result
        current_state = current_state_result["ret"]
    if not is_created:
        result["old_state"] = current_state
    resource_arn = current_state.get("db_cluster_parameter_group_arn")
    params_to_modify = {}
    plan_state = copy.deepcopy(current_state)

    if desired_state[
        "parameters"
    ] and not hub.tool.aws.state_comparison_utils.are_lists_identical(
        current_state["parameters"], desired_state["parameters"]
    ):
        params_to_modify["Parameters"] = desired_state["parameters"]
        plan_state["parameters"] = desired_state["parameters"]

    if params_to_modify:
        if ctx.get("test", False):
            if not is_created:
                result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                    resource_type=STATE_NAME, name=name
                )
        else:
            modify_db_cluster_parameter_group_result = (
                await hub.exec.boto3.client.neptune.modify_db_cluster_parameter_group(
                    ctx, DBClusterParameterGroupName=resource_id, **params_to_modify
                )
            )
            if not modify_db_cluster_parameter_group_result["result"]:
                result["comment"] = modify_db_cluster_parameter_group_result["comment"]
                result["result"] = False
                return result
            if not is_created:
                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.update_comment(
                    resource_type=STATE_NAME, name=name
                )

    # modify tags if it has changed
    old_tags = current_state.get("tags")
    new_tags = desired_state.get("tags")
    if new_tags is not None and new_tags != old_tags:
        # below code updates tags on AWS only if it is not a test run
        update_tags_ret = await hub.tool.aws.neptune.tag.update_tags(
            ctx=ctx,
            resource_arn=resource_arn,
            old_tags=old_tags,
            new_tags=new_tags,
        )
        if not update_tags_ret["result"]:
            result["comment"] = update_tags_ret["comment"]
            result["result"] = False
            hub.log.debug(f"Failed updating tags for {STATE_NAME} '{name}'")
            return result
        # update comment if tags were updated
        if not is_created:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type=STATE_NAME, name=name
            )
            result["comment"] += update_tags_ret["comment"]

        if ctx.get("test", False) and update_tags_ret["ret"] is not None:
            plan_state["tags"] = update_tags_ret["ret"]
            if not is_created:
                result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                    resource_type=STATE_NAME, name=resource_id
                )
                result["comment"] += update_tags_ret["comment"]

    # set new_state
    if ctx.get("test", False):
        result["new_state"] = plan_state
    else:
        current_db_cluster_parameter_group_result = (
            await hub.exec.aws.neptune.db_cluster_parameter_group.get(
                ctx=ctx, name=name, resource_id=resource_id
            )
        )
        if not (
            current_db_cluster_parameter_group_result
            and current_db_cluster_parameter_group_result["result"]
            and current_db_cluster_parameter_group_result["ret"]
        ):
            result["result"] = False
            result["comment"] += current_db_cluster_parameter_group_result["comment"]
            return result
        result["new_state"] = current_db_cluster_parameter_group_result["ret"]

    return result



"""Exec module for managing Organizations Policy."""


async def get(hub, ctx, name: str, resource_id: str):
    """Get Organization Policy from AWS.

    Provide Policy id as input.

    Args:
        name(str):
            An Idem name of the Organization policy.
        resource_id(str):
            AWS Organization policy ID

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The Organization Policy in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.organizations.policy.get name="idem_name" resource_id="resource_id"

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.organizations.policy.get(
                    ctx, name=name, resource_id=resource_id
                )
    """
    result = dict(comment=[], ret=None, result=True)
    tags = []
    before = await hub.exec.boto3.client.organizations.describe_policy(
        ctx, PolicyId=resource_id
    )
    if not before["result"]:
        if "PolicyNotFoundException" in str(before["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.organizations.policy", name=name
                )
            )
            result["comment"] += list(before["comment"])
            return result
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    resource_tags = await hub.exec.boto3.client.organizations.list_tags_for_resource(
        ctx, ResourceId=resource_id
    )
    if not resource_tags["result"]:
        result["result"] = False
        result["comment"] = resource_tags["comment"]
        return result
    if before["ret"]:
        tags = resource_tags["ret"].get("Tags", [])

    result[
        "ret"
    ] = hub.tool.aws.organizations.conversion_utils.convert_raw_policy_to_present(
        before["ret"]["Policy"], tags=tags if tags else None
    )
    return result



from typing import List


async def get(
    hub, ctx, *, name: str = None, filters: List = None, resource_id: str = None
):
    """
    Use an un-managed Instance-type as a data-source. Supply one of the inputs as the filter.

    Args:
        hub:
        ctx:
        name(Text): The name of the Idem state.
        resource_id(Text, optional): AWS Ec2 Instance type to identify the resource.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key.
        A complete list of filters can be found at https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_instance_types

    Request Syntax:
        [Idem-state-name]:
          aws.ec2.instance_type.search:
          - resource_id: 'Text'
          - filters:
            - name: 'string'
              values: 'list'
            - name: 'string'
              values: 'list'

        Examples:

            my-unmanaged-instance-type:
              exec.run:
                - path: aws.ec2.instance_type.get:
                - filters:
                  - name: instance-type
                    values:
                      - '*.nano'
                  - name: hypervisor
                    values:
                      - xen
                  - name: processor-info.supported-architecture
                    values:
                      - x86_64

            my_instance:
              aws.ec2.instance.present:
                - instance_type: ${aws.ec2.instance_type:my-unmanaged-instance-type:resource_id}
    """
    result = dict(comment=[], ret=None, result=True)

    # Perform validation on the parameters
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    result["result"] &= syntax_validation["result"]
    if not result["result"]:
        result["comment"].append(syntax_validation["comment"])
        return result

    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )

    # Get all instances that match the given filters
    ret = await hub.exec.boto3.client.ec2.describe_instance_types(
        ctx,
        Filters=boto3_filter,
        InstanceTypes=[resource_id] if resource_id else None,
    )
    result["result"] &= ret.result
    if not ret:
        result["comment"].append(ret.comment)
        return result

    # Get all the instance types that matched the filter
    instance_types = {i["InstanceType"] for i in ret.ret["InstanceTypes"]}

    # Check for null results
    if not instance_types:
        result["result"] = False
        result["comment"] += [
            f"Unable to find an aws.ec2.instance_type for '{name}' that matched the given filters"
        ]
        return result

    # Get the first instance from the results
    instance_type = next(iter(instance_types))

    # Add a comment if there were multiple results
    if len(instance_types) > 1:
        result["comment"] += [
            f"More than one aws.ec2.instance resource was found. Use resource '{instance_type}'"
        ]

    result["ret"] = dict(resource_id=instance_type)

    return result



"""State module for managing Amazon Cloudwatchlogs Resource Policy."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    policy_document: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Creates or updates a resource policy allowing other Amazon Web Services services to put log events to this account, such as Amazon Route 53.

    An account can have up to 10 resource policies per Amazon Web Services Region.

    Args:
        name(str):
            Name of the new policy. An Idem name of the resource

        policy_document(str): Details
            of the new policy, including the identity of the principal that is enabled to
            put logs to this account. This is formatted as a JSON string. This parameter is required.

        resource_id(str, Optional):
            AWS CloudWatchLogs resource policy name prefix

    Request Syntax:
        .. code-block:: sls

            [log_group_resource_policy_name]:
              aws.cloudwatchlogs.resource_policy.present:
               - name: 'string'
               - policy_document: 'string'
               - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

             resource_policy_is_present:
               aws.cloudwatchlogs.resource_policy.present:
                 - name: policy-tu62879
                 - policy_document: '{ "Version": "2012-10-17", "Statement": [ { "Sid": "Route53LogsToCloudWatchLogs",
                        "Effect": "Allow", "Principal": { "Service": [ "route53.amazonaws.com" ] }, "Action":
                        "logs:PutLogEvents", "Resource": "logArn", "Condition": { "ArnLike": { "aws:SourceArn":
                        "myRoute53ResourceArn" }, "StringEquals": { "aws:SourceAccount": "myAwsAccountId" } } } ] }'
                 - resource_id: policy-tu62879
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    plan_state = {}
    resource_policy_updated = False
    before = await hub.exec.aws.cloudwatchlogs.resource_policy.get(
        ctx, name=resource_id
    )
    if before["result"] and before["ret"]:
        result["comment"] = (f"'{name}' already exists",)
        result[
            "old_state"
        ] = hub.tool.aws.cloudwatchlogs.conversion_utils.convert_raw_log_group_resource_policy_to_present(
            ctx, raw_resource=before["ret"], idem_resource_name=resource_id
        )
        plan_state = copy.deepcopy(result["old_state"])
        if ctx.get("test", False):
            plan_state["policy_document"] = policy_document
        else:
            if policy_document != plan_state["policy_document"]:
                update_ret = await hub.exec.boto3.client.logs.put_resource_policy(
                    ctx, policyName=name, policyDocument=policy_document
                )
                if not update_ret["result"]:
                    result["comment"] = result["comment"] + update_ret["comment"]
                    result["result"] = False
                    return result
                resource_policy_updated = True
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "policy_document": policy_document,
                    "resource_id": name,
                },
            )
            result["comment"] = (
                f"Would create aws.cloudwatchlogs.resource_policy '{name}'",
            )
            return result
        try:
            ret = await hub.exec.boto3.client.logs.put_resource_policy(
                ctx, policyName=name, policyDocument=policy_document
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                return result
            resource_id = name
            result["comment"] = (
                f"Created aws.cloudwatchlogs.resource_policy '{name}'",
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif not (before and before["result"] and before["ret"]) or resource_policy_updated:
        before = await hub.exec.aws.cloudwatchlogs.resource_policy.get(
            ctx, name=resource_id
        )
        result["result"] = result["result"] and before["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + before["comment"]
        result[
            "new_state"
        ] = hub.tool.aws.cloudwatchlogs.conversion_utils.convert_raw_log_group_resource_policy_to_present(
            ctx, raw_resource=before["ret"], idem_resource_name=resource_id
        )
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""Deletes a resource policy from this account.

    This revokes the access of the identities in that policy to put log events to this account.

    Args:
        name(str):
            The name of the policy to be revoked. An Idem name of the resource

        resource_id(str, Optional):
            AWS CloudWatchLogs resource policy name prefix

    Request Syntax:
        .. code-block:: sls

            [log_group_resource_policy_name]:
              aws.cloudwatchlogs.resource_policy.absent:
              - name: 'string'
              - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_policy_is_absent:
              aws.cloudwatchlogs.resource_policy.absent:
                - name: policy-tu62879
                - resource_id: policy-tu62879
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.aws.cloudwatchlogs.resource_policy.get(
        ctx, name=resource_id
    )
    if not (before["result"] and before["ret"]):
        result["comment"] = (
            f"aws.cloudwatchlogs.resource_policy '{name}' already absent",
        )
    else:
        result[
            "old_state"
        ] = hub.tool.aws.cloudwatchlogs.conversion_utils.convert_raw_log_group_resource_policy_to_present(
            ctx, raw_resource=before["ret"], idem_resource_name=resource_id
        )
        if ctx.get("test", False):
            result["comment"] = result["comment"] + (
                f"Would delete aws.cloudwatchlogs.resource_policy '{name}'",
            )
            return result
        else:
            try:
                ret = await hub.exec.boto3.client.logs.delete_resource_policy(
                    ctx, policyName=name
                )
                result["result"] = ret["result"] and result["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + ret["comment"]
                    return result
                result["comment"] = result["comment"] + (
                    f"Deleted aws.cloudwatchlogs.resource_policy '{name}'",
                )
            except hub.tool.boto3.exception.ClientError as e:
                result["comment"] = result["comment"] + (
                    f"{e.__class__.__name__}: {e}",
                )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Lists the resource policies in this account.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.cloudwatchlogs.resource_policy
    """
    result = {}
    ret = await hub.exec.boto3.client.logs.describe_resource_policies(ctx)
    if not ret["result"]:
        hub.log.debug(
            f"Could not describe CloudWatchLogs Resource Policies {ret['comment']}"
        )
        return {}

    for resource in ret["ret"]["resourcePolicies"]:
        resource_id = resource.get("policyName")
        resource_translated = hub.tool.aws.cloudwatchlogs.conversion_utils.convert_raw_log_group_resource_policy_to_present(
            ctx, raw_resource=resource, idem_resource_name=resource_id
        )
        result[resource_id] = {
            "aws.cloudwatchlogs.resource_policy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name,
    resource_id: str,
) -> Dict:
    """
    Use an un-managed db_cluster_parameter_group as a data-source. Supply resource_id as a filter.

    Args:
        name(Text): The name of the Idem state.
        resource_id(Text):  AWS DB Cluster Parameter Group name.
    """

    result = dict(comment=[], ret=None, result=True)

    resource_ret = await hub.exec.boto3.client.rds.describe_db_cluster_parameter_groups(
        ctx, DBClusterParameterGroupName=resource_id
    )
    if not resource_ret["result"]:
        if "DBParameterGroupNotFound" in str(resource_ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.rds.db_cluster_parameter_group", name=name
                )
            )
            result["comment"] += list(resource_ret["comment"])
            return result
        result["comment"] += list(resource_ret["comment"])
        result["result"] = resource_ret["result"]
        return result

    if not resource_ret["ret"]["DBClusterParameterGroups"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.rds.db_cluster_parameter_group", name=name
            )
        )
        return result

    resource = resource_ret["ret"]["DBClusterParameterGroups"][0]
    resource_id = resource.get("DBClusterParameterGroupName")
    resource_arn = resource.get("DBClusterParameterGroupArn")
    ret_tags = await hub.exec.boto3.client.rds.list_tags_for_resource(
        ctx, ResourceName=resource_arn
    )

    if not ret_tags["result"]:
        result["comment"] += list(ret_tags["comment"])
        result["result"] = ret_tags["result"]
        return result

    tags_list = None
    if ret_tags["ret"]["TagList"]:
        tags_list = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            ret_tags["ret"]["TagList"]
        )

    ret_parameter = await hub.exec.boto3.client.rds.describe_db_cluster_parameters(
        ctx, DBClusterParameterGroupName=resource_id
    )

    if not ret_parameter["result"]:
        result["comment"] += list(ret_parameter["comment"])
        result["result"] = ret_parameter["result"]
        return result

    parameters = None
    if ret_parameter["ret"]["Parameters"]:
        parameters = ret_parameter["ret"]["Parameters"]

    result[
        "ret"
    ] = hub.tool.aws.rds.conversion_utils.convert_raw_db_cluster_parameter_group_to_present(
        raw_resource=resource,
        tags=tags_list,
        parameters=parameters,
    )

    return result


async def list_(
    hub,
    ctx,
    name,
) -> Dict:
    """
    Use an un-managed db_cluster_parameter_groups as a data-source.

    Args:
        name(string): The name of the Idem state.

    """
    result = dict(comment=[], ret=None, result=True)

    db_cluster_parameter_groups = []
    ret = await hub.exec.boto3.client.rds.describe_db_cluster_parameter_groups(ctx)
    if not ret["result"]:
        result["result"] = ret["result"]
        result["comment"] = ret["comment"]
        return result

    if not ret["ret"]["DBClusterParameterGroups"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.rds.db_cluster_parameter_group", name=name
            )
        )
        return result

    for resource in ret["ret"]["DBClusterParameterGroups"]:
        resource_id = resource.get("DBClusterParameterGroupName")
        resource_arn = resource.get("DBClusterParameterGroupArn")
        ret_tag = await hub.exec.boto3.client.rds.list_tags_for_resource(
            ctx, ResourceName=resource_arn
        )
        tags = []
        if ret_tag["result"]:
            tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                ret_tag.get("ret").get("TagList")
            )
        else:
            hub.log.debug(
                f"Could not get tags for db_cluster_parameter_groups {resource_id} {ret_tag['comment']}"
            )

        ret_parameter = await hub.exec.boto3.client.rds.describe_db_cluster_parameters(
            ctx, DBClusterParameterGroupName=resource_id
        )
        parameters = []
        if ret_parameter["result"]:
            parameters = ret_parameter["ret"]["Parameters"]
        else:
            hub.log.debug(
                f"Could not get parameters for db_cluster_parameter_groups {resource_id} {ret_parameter['comment']}"
            )

        converted_db_cluster_parameter_group = hub.tool.aws.rds.conversion_utils.convert_raw_db_cluster_parameter_group_to_present(
            raw_resource=resource,
            tags=tags,
            parameters=parameters,
        )
        db_cluster_parameter_groups.append(converted_db_cluster_parameter_group)

    result["ret"] = db_cluster_parameter_groups
    return result



"""State module for managing iam password policy."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    minimum_password_length: int = None,
    require_symbols: bool = False,
    require_numbers: bool = False,
    require_uppercase_characters: bool = False,
    require_lowercase_characters: bool = False,
    allow_users_to_change_password: bool = False,
    max_password_age: int = None,
    password_reuse_prevention: int = None,
    hard_expiry: bool = False,
) -> Dict[str, Any]:
    """Create/Update the password policy settings for the Amazon Web Services account.

    Args:
        name(str): An Idem name of the resource.

        minimum_password_length(int, optional): The minimum number of characters allowed in an IAM user password.
            If you do not specify a value for this parameter, then the operation uses the default value of 6.

        require_symbols(bool, optional): Specifies whether IAM user passwords must contain at least one of the following non-alphanumeric characters:
            ! @ # $ % ^ & * ( ) _ + - = [ ] { } | '
            If you do not specify a value for this parameter, then the operation uses the default value of false.
            The result is that passwords do not require at least one symbol character.

        require_numbers(bool, optional): Specifies whether IAM user passwords must contain at least one numeric character (0 to 9).
            If you do not specify a value for this parameter, then the operation uses the default value of false. The result is that passwords do not require at least one numeric character.

        require_uppercase_characters(bool, optional): Specifies whether IAM user passwords must contain at least one uppercase character from the ISO basic Latin alphabet (A to Z).
            If you do not specify a value for this parameter, then the operation uses the default value of false. The result is that passwords do not require at least one uppercase character.

        require_lowercase_characters(bool, optional): Specifies whether IAM user passwords must contain at least one lowercase character from the ISO basic Latin alphabet (a to z).
            If you do not specify a value for this parameter, then the operation uses the default value of false. The result is that passwords do not require at least one lowercase character.

        allow_users_to_change_password(bool, optional): Allows all IAM users in your account to use the Amazon Web Services Management Console to change their own passwords. For more information, see Permitting IAM users to change their own passwords in the IAM User Guide .
            If you do not specify a value for this parameter, then the operation uses the default value of false. The result is that IAM users in the account do not automatically have permissions to change their own password.

        max_password_age(int, optional): The number of days that an IAM user password is valid.
            If you do not specify a value for this parameter, then the operation uses the default value of 0. The result is that IAM user passwords never expire.

        password_reuse_prevention(int, optional): Specifies the number of previous passwords that IAM users are prevented from reusing.
            If you do not specify a value for this parameter, then the operation uses the default value of 0. The result is that IAM users are not prevented from reusing previous passwords.

        hard_expiry(bool, optional): Prevents IAM users who are accessing the account via the Amazon Web Services Management Console from setting a new console password after their password has expired. The IAM user cannot access the console until an administrator resets the password.
            If you do not specify a value for this parameter, then the operation uses the default value of false. The result is that IAM users can change their passwords after they expire and continue to sign in as the user.

    Request Syntax:
      .. code-block:: sls

        [password-policy]:
          aws.iam.password_policy.present:
          - minimum_password_length: 'string'
          - require_symbols: 'string'
          - require_numbers: 'string'
          - require_uppercase_characters: 'string'
          - require_lowercase_characters: 'string'
          - allow_users_to_change_password: 'string'
          - max_password_age: 'string'
          - password_reuse_prevention: 'string'
          - hard_expiry: 'string'

    Returns:
        Dict[str, str]

    Examples:
        .. code-block:: sls

            password-policy-011922870716:
              aws.iam.password_policy.present:
              - minimum_password_length: 6
              - require_symbols: true
              - require_numbers: true
              - require_uppercase_characters: true
              - require_lowercase_characters: true
              - allow_users_to_change_password: true
              - max_password_age: 90
              - password_reuse_prevention: 3
              - hard_expiry: true
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False
    plan_state = None

    before = await hub.exec.aws.iam.password_policy.get(ctx, name)
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    if before["ret"]:
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])

        policy_parameters = {
            "minimum_password_length": minimum_password_length,
            "require_symbols": require_symbols,
            "require_numbers": require_numbers,
            "require_uppercase_characters": require_uppercase_characters,
            "require_lowercase_characters": require_lowercase_characters,
            "allow_users_to_change_password": allow_users_to_change_password,
            "max_password_age": max_password_age,
            "password_reuse_prevention": password_reuse_prevention,
            "hard_expiry": hard_expiry,
        }

        update_ret = await hub.tool.aws.iam.password_policy.update_password_policy(
            ctx,
            name=name,
            before=result["old_state"],
            policy_parameters=policy_parameters,
        )
        result["comment"] += update_ret["comment"]
        resource_updated = resource_updated or bool(update_ret["ret"])
        if not update_ret["result"]:
            result["result"] = False
            return result

        if update_ret["ret"] and ctx.get("test", False):
            for modified_param in update_ret["ret"]:
                plan_state[modified_param] = update_ret["ret"][modified_param]
            result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.iam.password_policy", name=name
            )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "minimum_password_length": minimum_password_length,
                    "require_symbols": require_symbols,
                    "require_numbers": require_numbers,
                    "require_uppercase_characters": require_uppercase_characters,
                    "require_lowercase_characters": require_lowercase_characters,
                    "allow_users_to_change_password": allow_users_to_change_password,
                    "max_password_age": max_password_age,
                    "password_reuse_prevention": password_reuse_prevention,
                    "hard_expiry": hard_expiry,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.iam.password_policy", name=name
            )
            return result
        create_ret = await hub.exec.boto3.client.iam.update_account_password_policy(
            ctx,
            MinimumPasswordLength=minimum_password_length,
            RequireSymbols=require_symbols,
            RequireNumbers=require_numbers,
            RequireUppercaseCharacters=require_uppercase_characters,
            RequireLowercaseCharacters=require_lowercase_characters,
            AllowUsersToChangePassword=allow_users_to_change_password,
            MaxPasswordAge=max_password_age,
            PasswordReusePrevention=password_reuse_prevention,
            HardExpiry=hard_expiry,
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] = create_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.iam.password_policy", name=name
        )

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before["ret"]) or resource_updated:
            after_ret = await hub.exec.aws.iam.password_policy.get(ctx, name)
            if not after_ret["result"]:
                result["result"] = False
                result["comment"] += after_ret["comment"]
                return result
            result["new_state"] = after_ret["ret"]
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    """Deletes the password policy for the Amazon Web Services account.

    Args:
        name(str): Password policy name.

    Request Syntax:
        .. code-block:: sls

            password-policy:
              aws.iam.password_policy.absent:
              - name: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.iam.password_policy.absent:
              - name: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.aws.iam.password_policy.get(ctx, name)
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.password_policy", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.iam.password_policy", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.iam.delete_account_password_policy(ctx)
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.iam.password_policy", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Gets information about the iam password policy.

    Retrieves the password policy for the Amazon Web Services account.
    This tells you the complexity requirements and mandatory rotation periods for
    the IAM user passwords in your account.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

           $ idem describe aws.iam.password_policy

    """
    result = {}
    ret = await hub.exec.aws.iam.password_policy.get(ctx, "password_policy")
    if not ret["result"]:
        hub.log.debug(f"Could not describe account password policy {ret['comment']}")
        return {}
    resource_key = "password-policy"
    result[resource_key] = {
        "aws.iam.password_policy.present": [
            {parameter_key: parameter_value}
            for parameter_key, parameter_value in ret["ret"].items()
        ]
    }

    return result



"""State module for AWS neptune DB instance."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

STATE_NAME = "aws.neptune.db_instance"


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    *,
    db_instance_class: str,
    engine: str,
    db_name: str = None,
    allocated_storage: int = None,
    master_username: str = None,
    master_user_password: str = None,
    db_security_groups: List[str] = None,
    vpc_security_group_ids: List[str] = None,
    availability_zone: str = None,
    db_subnet_group_name: str = None,
    preferred_maintenance_window: str = None,
    db_parameter_group_name: str = None,
    backup_retention_period: int = None,
    preferred_backup_window: str = None,
    port: int = None,
    multi_az: bool = None,
    engine_version: str = None,
    auto_minor_version_upgrade: bool = None,
    license_model: str = None,
    iops: int = None,
    option_group_name: str = None,
    character_set_name: str = None,
    publicly_accessible: bool = None,
    tags: Dict[str, Any],
    db_cluster_identifier: str = None,
    storage_type: str = None,
    tde_credential_arn: str = None,
    tde_credential_password: str = None,
    storage_encrypted: bool = None,
    kms_key_id: str = None,
    domain: str = None,
    copy_tags_to_snapshot: bool = None,
    monitoring_interval: int = None,
    monitoring_role_arn: str = None,
    domain_iam_role_name: str = None,
    promotion_tier: int = None,
    timezone: str = None,
    enable_iam_database_authentication: bool = None,
    enable_performance_insights: bool = None,
    performance_insights_kms_key_id: str = None,
    enable_cloudwatch_logs_exports: List[str] = None,
    deletion_protection: bool = None,
    apply_immediately: bool = None,
    allow_major_version_upgrade: bool = None,
    ca_certificate_identifier: str = None,
    cloudwatch_logs_export_configuration: make_dataclass(
        """Cloudwatch log export configuration dataclass."""
        "CloudwatchLogsExportConfiguration",
        [
            ("EnableLogTypes", List[str], field(default=None)),
            ("DisableLogTypes", List[str], field(default=None)),
        ],
    ) = None,
    timeout: make_dataclass(
        """Timeout configuration dataclass.""" "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=15)),
                        ("max_attempts", int, field(default=40)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=15)),
                        ("max_attempts", int, field(default=40)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """Creates a new DB instance.

    Args:
        name(str):
            An Idem name of the resource. This is also DB instance identifier during resource creation. This
            parameter is stored as a lowercase string.
            Constraints:   Must contain from 1 to 63 letters, numbers, or hyphens.   First character must be a letter.
            Cannot end with a hyphen or contain two consecutive hyphens.   Example: mydbinstance.
        resource_id(str, Optional):
            AWS Neptune DBInstanceIdentifier to identify the resource. Defaults to None.
        db_name(str, Optional):
            Not supported. Defaults to None.
        allocated_storage(int, Optional):
            Not supported by Neptune. Defaults to None.
        db_instance_class(str):
            The compute and memory capacity of the DB instance, for example, db.m4.large. Not all DB
            instance classes are available in all Amazon Regions.
        engine(str):
            The name of the database engine to be used for this instance. Valid Values: neptune.
        master_username(str, Optional):
            Not supported by Neptune. Defaults to None.
        master_user_password(str, Optional):
            Not supported by Neptune. Defaults to None.
        db_security_groups(list[str], Optional):
            A list of DB security groups to associate with this DB instance. Default: The default DB
            security group for the database engine. Defaults to None.
        vpc_security_group_ids(list[str], Optional):
            A list of EC2 VPC security groups to associate with this DB instance. Not applicable. The
            associated list of EC2 VPC security groups is managed by the DB cluster. For more information,
            see CreateDBCluster. Default: The default EC2 VPC security group for the DB subnet group's VPC. Defaults to None.
        availability_zone(str, Optional):
            The EC2 Availability Zone that the DB instance is created in Default: A random, system-chosen
            Availability Zone in the endpoint's Amazon Region.  Example: us-east-1d   Constraint: The
            AvailabilityZone parameter can't be specified if the MultiAZ parameter is set to true. The
            specified Availability Zone must be in the same Amazon Region as the current endpoint. Defaults to None.
        db_subnet_group_name(str, Optional):
            A DB subnet group to associate with this DB instance. If there is no DB subnet group, then it is
            a non-VPC DB instance. Defaults to None.
        preferred_maintenance_window(str, Optional):
            The time range each week during which system maintenance can occur, in Universal Coordinated
            Time (UTC).  Format: ``ddd:hh24:mi-ddd:hh24:mi``  The default is a 30-minute window selected at
            random from an 8-hour block of time for each Amazon Region, occurring on a random day of the
            week. Valid Days: Mon, Tue, Wed, Thu, Fri, Sat, Sun. Constraints: Minimum 30-minute window. Defaults to None.
        db_parameter_group_name(str, Optional):
            The name of the DB parameter group to associate with this DB instance. If this argument is
            omitted, the default DBParameterGroup for the specified engine is used. Constraints:   Must be 1
            to 255 letters, numbers, or hyphens.   First character must be a letter   Cannot end with a
            hyphen or contain two consecutive hyphens. Defaults to None.
        backup_retention_period(int, Optional):
            The number of days for which automated backups are retained. Not applicable. The retention
            period for automated backups is managed by the DB cluster. For more information, see
            CreateDBCluster. Default: 1 Constraints:   Must be a value from 0 to 35   Cannot be set to 0 if
            the DB instance is a source to Read Replicas. Defaults to None.
        preferred_backup_window(str, Optional):
            The daily time range during which automated backups are created. Not applicable. The daily time
            range for creating automated backups is managed by the DB cluster. For more information, see
            CreateDBCluster. Defaults to None.
        port(int, Optional):
            The port number on which the database accepts connections. Not applicable. The port is managed
            by the DB cluster. For more information, see CreateDBCluster.  Default: 8182  Type: Integer. Defaults to None.
        multi_az(bool, Optional):
            Specifies if the DB instance is a Multi-AZ deployment. You can't set the AvailabilityZone
            parameter if the MultiAZ parameter is set to true. Defaults to None.
        engine_version(str, Optional):
            The version number of the database engine to use. Currently, setting this parameter has no
            effect. Defaults to None.
        auto_minor_version_upgrade(bool, Optional):
            Indicates that minor engine upgrades are applied automatically to the DB instance during the
            maintenance window. Default: true. Defaults to None.
        license_model(str, Optional):
            License model information for this DB instance.  Valid values: license-included | bring-your-
            own-license | general-public-license. Defaults to None.
        iops(int, Optional):
            The amount of Provisioned IOPS (input/output operations per second) to be initially allocated
            for the DB instance. Defaults to None.
        option_group_name(str, Optional):
            (Not supported by Neptune). Defaults to None.
        character_set_name(str, Optional):
            (Not supported by Neptune). Defaults to None.
        publicly_accessible(bool, Optional):
            This flag should no longer be used. Defaults to None.
        tags(dict, Optional):
            Dict in the format of {tag-key: tag-value} to associate with the VPC.
            Each tag consists of a key name and an associated value. Defaults to None.

            * Key (str, Optional):
                The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.

            * Value(str, Optional):
                The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.
        db_cluster_identifier(str, Optional):
            The identifier of the DB cluster that the instance will belong to. For information on creating a
            DB cluster, see CreateDBCluster. Type: String. Defaults to None.
        storage_type(str, Optional):
            Specifies the storage type to be associated with the DB instance. Not applicable. Storage is
            managed by the DB Cluster. Defaults to None.
        tde_credential_arn(str, Optional):
            The ARN from the key store with which to associate the instance for TDE encryption. Defaults to None.
        tde_credential_password(str, Optional):
            The password for the given ARN from the key store in order to access the device. Defaults to None.
        storage_encrypted(bool, Optional):
            Specifies whether the DB instance is encrypted. Not applicable. The encryption for DB instances
            is managed by the DB cluster. For more information, see CreateDBCluster. Default: false. Defaults to None.
        kms_key_id(str, Optional):
            The Amazon KMS key identifier for an encrypted DB instance. The KMS key identifier is the Amazon
            Resource Name (ARN) for the KMS encryption key. If you are creating a DB instance with the same
            Amazon account that owns the KMS encryption key used to encrypt the new DB instance, then you
            can use the KMS key alias instead of the ARN for the KM encryption key. Not applicable. The KMS
            key identifier is managed by the DB cluster. For more information, see CreateDBCluster. If the
            StorageEncrypted parameter is true, and you do not specify a value for the KmsKeyId parameter,
            then Amazon Neptune will use your default encryption key. Amazon KMS creates the default
            encryption key for your Amazon account. Your Amazon account has a different default encryption
            key for each Amazon Region. Defaults to None.
        domain(str, Optional):
            Specify the Active Directory Domain to create the instance in. Defaults to None.
        copy_tags_to_snapshot(bool, Optional):
            True to copy all tags from the DB instance to snapshots of the DB instance, and otherwise false.
            The default is false. Defaults to None.
        monitoring_interval(int, Optional):
            The interval, in seconds, between points when Enhanced Monitoring metrics are collected for the
            DB instance. To disable collecting Enhanced Monitoring metrics, specify 0. The default is 0. If
            MonitoringRoleArn is specified, then you must also set MonitoringInterval to a value other than
            0. Valid Values: 0, 1, 5, 10, 15, 30, 60. Defaults to None.
        monitoring_role_arn(str, Optional):
            The ARN for the IAM role that permits Neptune to send enhanced monitoring metrics to Amazon
            CloudWatch Logs. For example, arn:aws:iam:123456789012:role/emaccess. If MonitoringInterval is
            set to a value other than 0, then you must supply a MonitoringRoleArn value. Defaults to None.
        domain_iam_role_name(str, Optional):
            Specify the name of the IAM role to be used when making API calls to the Directory Service. Defaults to None.
        promotion_tier(int, Optional):
            A value that specifies the order in which an Read Replica is promoted to the primary instance
            after a failure of the existing primary instance.  Default: 1 Valid Values: 0 - 15. Defaults to None.
        timezone(str, Optional):
            The time zone of the DB instance. Defaults to None.
        enable_iam_database_authentication(bool, Optional):
            Not supported by Neptune (ignored). Defaults to None.
        enable_performance_insights(bool, Optional):
            (Not supported by Neptune). Defaults to None.
        performance_insights_kms_key_id(str, Optional):
            (Not supported by Neptune). Defaults to None.
        enable_cloudwatch_logs_exports(list[str], Optional):
            The list of log types that need to be enabled for exporting to CloudWatch Logs. Defaults to None.
        deletion_protection(bool, Optional):
            A value that indicates whether the DB instance has deletion protection enabled. The database
            can't be deleted when deletion protection is enabled. By default, deletion protection is
            disabled. See Deleting a DB Instance. DB instances in a DB cluster can be deleted even when
            deletion protection is enabled in their parent DB cluster. Defaults to None.
        apply_immediately(bool, Optional):
            Specifies whether the modifications in this request and any pending modifications are asynchronously applied as soon as possible, regardless of the PreferredMaintenanceWindow setting for the DB instance.
            If this parameter is set to false , changes to the DB instance are applied during the next maintenance window. Some parameter changes can cause an outage and are applied on the next call to RebootDBInstance , or the next failure reboot.
            Default: false
        allow_major_version_upgrade(bool, Optional):
            Indicates that major version upgrades are allowed. Changing this parameter doesn't result in an outage and the change is asynchronously applied as soon as possible.
        ca_certificate_identifier(str, Optional):
            Indicates the certificate that needs to be associated with the instance.
        cloudwatch_logs_export_configuration(dict[str, Any], Optional):
            The configuration setting for the log types to be enabled for export to CloudWatch Logs for a specific DB instance.

            * EnableLogTypes (list[str], Optional):
                The list of log types to enable.

            * DisableLogTypes(list[str], Optional):
                The list of log types to disable.
        timeout(dict, Optional):
            Timeout configuration for create/update of AWS DB Instance.

            * create (*dict, Optional*):
                Timeout configuration for creating DB Cluster.

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

            * update(*dict, Optional*):
                Timeout configuration for updating DB instance.

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        .. code-block:: sls

            [idem_test_aws_neptune_db_instance]:
                  aws.neptune.db_instance.present:
                      - name: 'string'
                      - resource_id: 'string'
                      - db_instance_identifier: 'string'
                      - db_instance_class: 'string'
                      - engine: 'string'
                      - allocated_storage: 'integer'
                      - master_username: 'string'
                      - db_security_groups:
                            - 'string'
                            - 'string'
                      - availability_zone: 'string'
                      - preferred_maintenance_window: 'string'
                      - backup_retention_period: 'integer'
                      - preferred_backup_window: 'string'
                      - multi_az: 'boolean'
                      - engine_version: 'string'
                      - auto_minor_version_upgrade: 'boolean'
                      - license_model: 'string'
                      - publicly_accessible: 'boolean'
                      - db_cluster_identifier: 'string'
                      - storage_type: 'string'
                      - storage_encrypted: 'boolean'
                      - kms_key_id: 'string'
                      - copy_tags_to_snapshot: 'boolean'
                      - monitoring_interval: 'integer'
                      - promotion_tier: 'integer'
                      - enable_iam_database_authentication: 'boolean'
                      - enable_performance_insights: 'boolean'
                      - enable_cloudwatch_logs_exports:
                        - 'string'
                      - db_instance_arn: 'string'
                      - deletion_protection: 'boolean'
                      - vpc_security_group_ids:
                        - 'string'
                      - db_parameter_group_name: 'string'
                      - option_group_name: 'string'
                      - db_subnet_group_name: 'string'

    Returns:
        Dict[str, Dict[str,Any]]

    Examples:
        .. code-block:: sls

            database-1-instance-1:
              aws.neptune.db_instance.present:
                  - name: database-1-instance-1
                  - resource_id: database-1-instance-1
                  - db_instance_identifier: database-1-instance-1
                  - db_instance_class: db.t3.medium
                  - engine: neptune
                  - allocated_storage: 1
                  - master_username: admin
                  - db_security_groups: []
                  - availability_zone: ca-central-1d
                  - preferred_maintenance_window: sun:03:07-sun:03:37
                  - backup_retention_period: 1
                  - preferred_backup_window: 06:18-06:48
                  - multi_az: false
                  - engine_version: 1.1.1.0
                  - auto_minor_version_upgrade: true
                  - license_model: amazon-license
                  - publicly_accessible: false
                  - db_cluster_identifier: database-1
                  - storage_type: aurora
                  - storage_encrypted: true
                  - kms_key_id: arn:aws:kms:ca-central-1:746014882121:key/2a4a587a-04fb-4c3b-b4d5-c3f25fb7f69f
                  - copy_tags_to_snapshot: false
                  - monitoring_interval: 0
                  - promotion_tier: 1
                  - enable_iam_database_authentication: false
                  - enable_performance_insights: false
                  - enable_cloudwatch_logs_exports:
                    - audit
                  - db_instance_arn: arn:aws:rds:ca-central-1:746014882121:db:database-1-instance-1
                  - deletion_protection: false
                  - vpc_security_group_ids:
                    - sg-ad5a2ac5
                  - db_parameter_group_name: default.neptune1
                  - option_group_name: default:neptune-1-1
                  - db_subnet_group_name: default
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    # when creating, below is the desired state of resource
    desired_state = {
        "name": name,
        "resource_id": name,
        "db_instance_class": db_instance_class,
        "engine": engine,
        "db_name": db_name,
        "allocated_storage": allocated_storage,
        "master_username": master_username,
        "master_user_password": master_user_password,
        "db_security_groups": db_security_groups,
        "vpc_security_group_ids": vpc_security_group_ids,
        "availability_zone": availability_zone,
        "db_subnet_group_name": db_subnet_group_name,
        "preferred_maintenance_window": preferred_maintenance_window,
        "db_parameter_group_name": db_parameter_group_name,
        "backup_retention_period": backup_retention_period,
        "preferred_backup_window": preferred_backup_window,
        "port": port,
        "multi_az": multi_az,
        "engine_version": engine_version,
        "auto_minor_version_upgrade": auto_minor_version_upgrade,
        "license_model": license_model,
        "iops": iops,
        "option_group_name": option_group_name,
        "character_set_name": character_set_name,
        "publicly_accessible": publicly_accessible,
        "tags": tags,
        "db_cluster_identifier": db_cluster_identifier,
        "storage_type": storage_type,
        "tde_credential_arn": tde_credential_arn,
        "tde_credential_password": tde_credential_password,
        "storage_encrypted": storage_encrypted,
        "kms_key_id": kms_key_id,
        "domain": domain,
        "copy_tags_to_snapshot": copy_tags_to_snapshot,
        "monitoring_interval": monitoring_interval,
        "monitoring_role_arn": monitoring_role_arn,
        "domain_iam_role_name": domain_iam_role_name,
        "promotion_tier": promotion_tier,
        "timezone": timezone,
        "enable_iam_database_authentication": enable_iam_database_authentication,
        "enable_performance_insights": enable_performance_insights,
        "performance_insights_kms_key_id": performance_insights_kms_key_id,
        "enable_cloudwatch_logs_exports": enable_cloudwatch_logs_exports,
        "deletion_protection": deletion_protection,
    }
    # set a flag to keep track if the resource is newly created
    is_created = False

    if not resource_id:
        # if there is no resource_id, it is expected that we create the resource
        # if this is test run, just generate the new state and set in result
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={}, desired_state=desired_state
            )
            resource_id = desired_state["resource_id"]
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type=STATE_NAME, name=name
            )
        else:
            # if this is not a test run, create the real instance
            create_db_instance_result = (
                await hub.exec.boto3.client.neptune.create_db_instance(
                    ctx,
                    DBName=db_name,
                    DBInstanceIdentifier=name,
                    AllocatedStorage=allocated_storage,
                    DBInstanceClass=db_instance_class,
                    Engine=engine,
                    MasterUsername=master_username,
                    MasterUserPassword=master_user_password,
                    DBSecurityGroups=db_security_groups,
                    VpcSecurityGroupIds=vpc_security_group_ids,
                    AvailabilityZone=availability_zone,
                    DBSubnetGroupName=db_subnet_group_name,
                    PreferredMaintenanceWindow=preferred_maintenance_window,
                    DBParameterGroupName=db_parameter_group_name,
                    BackupRetentionPeriod=backup_retention_period,
                    PreferredBackupWindow=preferred_backup_window,
                    Port=port,
                    MultiAZ=multi_az,
                    EngineVersion=engine_version,
                    AutoMinorVersionUpgrade=auto_minor_version_upgrade,
                    LicenseModel=license_model,
                    Iops=iops,
                    OptionGroupName=option_group_name,
                    CharacterSetName=character_set_name,
                    PubliclyAccessible=publicly_accessible,
                    Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                    DBClusterIdentifier=db_cluster_identifier,
                    StorageType=storage_type,
                    TdeCredentialArn=tde_credential_arn,
                    TdeCredentialPassword=tde_credential_password,
                    StorageEncrypted=storage_encrypted,
                    KmsKeyId=kms_key_id,
                    Domain=domain,
                    CopyTagsToSnapshot=copy_tags_to_snapshot,
                    MonitoringInterval=monitoring_interval,
                    MonitoringRoleArn=monitoring_role_arn,
                    DomainIAMRoleName=domain_iam_role_name,
                    PromotionTier=promotion_tier,
                    Timezone=timezone,
                    EnableIAMDatabaseAuthentication=enable_iam_database_authentication,
                    EnablePerformanceInsights=enable_performance_insights,
                    PerformanceInsightsKMSKeyId=performance_insights_kms_key_id,
                    EnableCloudwatchLogsExports=enable_cloudwatch_logs_exports,
                    DeletionProtection=deletion_protection,
                )
            )
            # if create failed, fail the request
            if not create_db_instance_result["result"]:
                result["comment"] = create_db_instance_result["comment"]
                result["result"] = False
                return result

            # using the resource_id, wait till db instance is created
            resource_id = create_db_instance_result["ret"]["DBInstance"][
                "DBInstanceIdentifier"
            ]
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=15,
                default_max_attempts=40,
                timeout_config=timeout.get("create") if timeout else None,
            )
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "neptune",
                    "db_instance_available",
                    DBInstanceIdentifier=resource_id,
                    WaiterConfig=waiter_config,
                )
                result["comment"] = hub.tool.aws.comment_utils.create_comment(
                    resource_type=STATE_NAME, name=name
                )
            except Exception as e:
                result["comment"] += (str(e),)
                result["result"] = False
                return result
        is_created = True

    # neptune db_instance has some parameters that can only be modified, so if
    # the user specifies modification parameters during creation of new db instance, idem will
    # create the instance and also modify it so that desired state is obtained
    desired_state["apply_immediately"] = apply_immediately
    desired_state[
        "cloudwatch_logs_export_configuration"
    ] = cloudwatch_logs_export_configuration
    desired_state["ca_certificate_identifier"] = ca_certificate_identifier
    desired_state["allow_major_version_upgrade"] = allow_major_version_upgrade

    update_db_instance_result = await hub.tool.aws.neptune.db_instance.update(
        ctx=ctx,
        name=name,
        resource_id=resource_id,
        desired_state=desired_state,
        is_created=is_created,
        timeout=timeout,
    )
    if not (
        update_db_instance_result
        and update_db_instance_result["result"]
        and update_db_instance_result["new_state"]
    ):
        result["result"] = False
        result["comment"] = update_db_instance_result["comment"]
        return result

    result["old_state"] = update_db_instance_result["old_state"]
    result["new_state"] = update_db_instance_result["new_state"]
    # update the result comment if the resource is not being created
    if not is_created:
        result["comment"] = update_db_instance_result["comment"]

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str,
    skip_final_snapshot: bool = None,
    final_db_snapshot_identifier: str = None,
    timeout: make_dataclass(
        """Timeout configuration.""" "Timeout",
        [
            (
                "delete",
                make_dataclass(
                    "DeleteTimeout",
                    [
                        ("delay", int, field(default=15)),
                        ("max_attempts", int, field(default=40)),
                    ],
                ),
                field(default=None),
            )
        ],
    ) = None,
) -> Dict[str, Any]:
    """Deletes a previously provisioned DB instance.

    When you delete a DB instance, all automated backups for that instance are deleted and can't be recovered.
    Manual DB snapshots of the DB instance to be deleted by DeleteDBInstance are not deleted.  If you request a final
    DB snapshot the status of the Amazon Neptune DB instance is deleting until the DB snapshot is created.
    The API action DescribeDBInstance is used to monitor the status of this operation. The action can't be canceled or
    reverted once submitted. Note that when a DB instance is in a failure state and has a status of failed,
    incompatible-restore, or incompatible-network, you can only delete it when the SkipFinalSnapshot parameter is set
    to true. You can't delete a DB instance if it is the only instance in the DB cluster, or if it has deletion
    protection enabled.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str):
            AWS Neptune DBInstanceIdentifier to identify the resource. This parameter isn't case-
            sensitive. Idem automatically considers the resource being absent if resource_id is not specified.
            Constraints:   Must match the name of an existing DB instance.
        skip_final_snapshot(bool, Optional):
            Determines whether a final DB snapshot is created before the DB instance is deleted. If true is
            specified, no DBSnapshot is created. If false is specified, a DB snapshot is created before the
            DB instance is deleted. Note that when a DB instance is in a failure state and has a status of
            'failed', 'incompatible-restore', or 'incompatible-network', it can only be deleted when the
            SkipFinalSnapshot parameter is set to "true". Specify true when deleting a Read Replica.  The
            FinalDBSnapshotIdentifier parameter must be specified if SkipFinalSnapshot is false.  Default:
            false. Defaults to None.
        final_db_snapshot_identifier(str, Optional):
            The DBSnapshotIdentifier of the new DBSnapshot created when SkipFinalSnapshot is set to false.
            Specifying this parameter and also setting the SkipFinalShapshot parameter to true results in an
            error.  Constraints:   Must be 1 to 255 letters or numbers.   First character must be a letter
            Cannot end with a hyphen or contain two consecutive hyphens   Cannot be specified when deleting
            a Read Replica. Defaults to None.
        timeout(dict, Optional):
            Timeout configuration for create/update of AWS DB Instance.

            * delete (*dict, Optional*):
                Timeout configuration for deleting DB Instance.

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        .. code-block:: sls

            [idem_test_aws_neptune_db_instance]:
                aws.neptune.db_instance.absent:
                    - name: 'string'
                    - resource_id: 'string'
                    - skip_final_snapshot: 'boolean'
                    - final_db_snapshot_identifier: 'string'
                    - timeout:
                        delete:
                            delay: int
                            max_attempts: int

    Returns:
        Dict[str, Dict[str,Any]]

    Examples:
        .. code-block:: sls

            database-1-instance-1:
                aws.neptune.db_instance.absent:
                    - name: database-1-instance-1
                    - resource_id: database-1-instance-1
                    - skip_final_snapshot: True
    """
    already_absent_msg = hub.tool.aws.comment_utils.already_absent_comment(
        resource_type=STATE_NAME, name=name
    )
    result = dict(
        comment=already_absent_msg,
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )
    if not resource_id:
        return result
    before = await hub.exec.aws.neptune.db_instance.get(
        ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if not before["ret"]:
        return result
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type=STATE_NAME, name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.neptune.delete_db_instance(
            ctx,
            DBInstanceIdentifier=resource_id,
            SkipFinalSnapshot=skip_final_snapshot,
            FinalDBSnapshotIdentifier=final_db_snapshot_identifier,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type=STATE_NAME, name=name
        )

        # Waiting for the DBInstance to Delete
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=15,
            default_max_attempts=40,
            timeout_config=timeout.get("delete") if timeout else None,
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "neptune",
                "db_instance_deleted",
                DBInstanceIdentifier=resource_id,
                WaiterConfig=waiter_config,
            )
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type=STATE_NAME, name=name
            )
        except Exception as e:
            result["comment"] = (str(e),)
            result["result"] = False

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns information about provisioned instances, and supports pagination.  This operation can also return
    information for Amazon RDS instances and Amazon DocDB instances.

    .. note::
        Following sensitive parameters are part of present but are not available as part of describe:
        * master_user_password
        * tde_credential_password

        Following parameter is not part of present and is ignored from describe,
        idem ensures that the status is in ready state before declaring the resource present:
        * db_instance_status

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.neptune.db_instance
    """
    result = {}
    ret = await hub.exec.boto3.client.neptune.describe_db_instances(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe db_instance {ret['comment']}")
        return {}

    for db_instance in ret["ret"]["DBInstances"]:
        # Including fields to match the 'present' function parameters
        resource_id = db_instance.get("DBInstanceIdentifier")
        resource_arn = db_instance.get("DBInstanceArn")
        tags = await hub.tool.aws.neptune.tag.get_tags_for_resource(
            ctx, resource_arn=resource_arn
        )
        if not tags["result"]:
            hub.log.warning(
                f"Failed listing tags for aws.neptune.db_instance '{resource_id}'"
                f"Describe will skip this aws.neptune.db_instance and continue. "
            )
            continue
        tags = tags["ret"]
        resource_translated = await hub.tool.aws.neptune.conversion_utils.convert_raw_db_instance_to_present(
            db_instance, idem_resource_name=resource_id, tags=tags
        )
        result[resource_id] = {
            "aws.neptune.db_instance.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



"""Exec module for managing Amazon IAM Policies."""
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict:
    """Retrieves the specified AWS IAM Policy.

    Args:
        name(str):
            The name of the IAM Policy.
        resource_id(str, Optional):
            The Amazon Resource Name (ARN) of the IAM policy in Amazon Web Services. If not supplied, ``name`` will be used to get the resource.

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The IAM Policy in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.iam.policy.get name="idem_name" resource_id="resource_id"

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.iam.policy.get(
                    ctx, name=name, resource_id=resource_id
                )
    """
    result = dict(comment=[], ret=None, result=True)

    if not resource_id:
        ret = await hub.exec.boto3.client.iam.list_policies(ctx)
        if not ret["result"]:
            result["comment"] += list(ret["comment"])
            result["result"] = False
            return result
        for ret_policy in ret["ret"]["Policies"]:
            if ret_policy["PolicyName"] == name:
                resource_id = ret_policy["Arn"]
                continue
        if not resource_id:
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.iam.policy", name=name
                )
            )
            return result

    ret = await hub.exec.boto3.client.iam.get_policy(ctx, PolicyArn=resource_id)
    if not ret["result"]:
        if "NoSuchEntityException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.iam.policy", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    policy = ret["ret"]["Policy"]
    ret_get_policy_version = await hub.exec.boto3.client.iam.get_policy_version(
        ctx, PolicyArn=policy["Arn"], VersionId=policy["DefaultVersionId"]
    )
    if not ret_get_policy_version["result"]:
        result["comment"] += list(ret_get_policy_version)
        result["result"] = False
        return result
    policy["Document"] = ret_get_policy_version["ret"]["PolicyVersion"].get("Document")

    result[
        "ret"
    ] = await hub.tool.aws.iam.conversion_utils.convert_raw_policy_to_present(
        ctx, raw_resource=policy
    )

    return result


async def list_(
    hub,
    ctx,
    scope: str = None,
    only_attached: bool = None,
    path_prefix: str = None,
    policy_usage_filter: str = None,
) -> Dict:
    """Lists AWS IAM Policies.

    Arg:
        scope(str, Optional):
            The scope to use for filtering the results. To list only Amazon Web Services managed policies, set ``scope`` to ``AWS``.
            To list only the customer managed policies in your Amazon Web Services account, set ``scope`` to ``Local``.
            This parameter is optional. If it is not included, or if it is set to ``All``, all policies are returned.
        only_attached (bool, Optional):
            A flag to filter the results to only the attached policies. When ``True``, the returned list contains only the policies
            that are attached to an IAM user, group, or role. When ``False``, or when the parameter is not included, all policies are returned.
        path_prefix (str, Optional):
            The path prefix for filtering the results. This parameter is optional. If it is not included, it defaults to a slash (``/``),
            listing all policies. This parameter allows (through its regex pattern) a string of characters consisting of either a forward
            slash (``/``) by itself or a string that must begin and end with forward slashes. In addition, it can contain any ASCII character
            from the ``!`` (``\u0021``) through the ``DEL`` character (``\u007F``), including most punctuation characters, digits, and upper
            and lowercased letters.
        policy_usage_filter(str, Optional):
            The policy usage method to use for filtering the results. To list only permissions policies, set the value to ``PermissionsPolicy``.
            To list only the policies used to set permissions boundaries, set the value to ``PermissionsBoundary``. This parameter is optional.
            If it is not included, all policies are returned.

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The IAM Policies in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.iam.policy.list scope="ALL"

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, scope, **kwargs):
                ret = await hub.exec.aws.iam.policy.list(
                    ctx, scope=scope
                )
    """
    result = dict(comment=[], ret=[], result=True)

    ret = await hub.exec.boto3.client.iam.list_policies(
        ctx=ctx,
        Scope=scope,
        OnlyAttached=only_attached,
        PathPrefix=path_prefix,
        PolicyUsageFilter=policy_usage_filter,
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Policies"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.iam.policy", name=scope
            )
        )
        return result

    for policy in ret["ret"]["Policies"]:
        get_ret = await hub.exec.aws.iam.policy.get(
            ctx, name=policy["PolicyName"], resource_id=policy["Arn"]
        )
        if not get_ret["result"]:
            result["comment"] += list(get_ret["comment"])
            result["result"] = False
            return result
        if get_ret["ret"]:
            result["ret"].append(get_ret["ret"])

    return result



from typing import Any
from typing import Dict


async def get_ssh_key_if_attached_to_user(
    hub, ctx, user_name: str, ssh_public_key_id: str
) -> Dict[str, Any]:
    """
    Check if ssh_public_key_id is already associated to this user. If already associated
    return the metadata of ssh public key. If not associated return {} in ret.

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        user_name: The name (friendly name, not ARN) of the IAM user
        ssh_public_key_id(Text): The SSH public key ID

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}
    """
    result = dict(comment=(), result=True, ret=None)
    user_ssh_public_keys_list = await hub.exec.boto3.client.iam.list_ssh_public_keys(
        ctx, UserName=user_name
    )
    if user_ssh_public_keys_list.get("result"):
        attached_user_ssh_public_keys_list = user_ssh_public_keys_list["ret"].get(
            "SSHPublicKeys"
        )
        if attached_user_ssh_public_keys_list:
            for ssh_public_key in attached_user_ssh_public_keys_list:
                if ssh_public_key.get("SSHPublicKeyId") == ssh_public_key_id:
                    result["ret"] = ssh_public_key
                    break
    else:
        result["comment"] = user_ssh_public_keys_list["comment"]
        result["result"] = False
    return result



from typing import Any
from typing import Dict
from typing import List

from dict_tools import data


async def update(
    hub, ctx, resource_id: str, before: Dict[str, Any], role: str, rules: List
):
    """Updates the replication configuration for the S3 bucket.

    Update the replication configuration for provided bucket.

    Args:
        hub: required for functions in hub.
        ctx: context.
        resource_id: AWS S3 bucket name.
        before(Dict): Existing resource parameters in Amazon Web Services.
        role (str):
            The Amazon Resource Name (ARN) of the Identity and Access Management (IAM) role that Amazon S3
            assumes when replicating objects.
        rules (list):
            A container for one or more replication rules. A replication configuration must have at least
            one rule and can contain a maximum of 1,000 rules.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}

    """
    result = dict(comment=(), result=True, ret=None)
    update_payload = {}
    payload = {}

    if role and role != before.get("role"):
        update_payload["Role"] = role
        payload["Role"] = role
    else:
        payload["Role"] = before.get("role")

    if rules:
        before_rule_map = {}
        rule_map = {}

        for rule in before["rules"]:
            before_rule_map[rule["ID"]] = rule

        for rule in rules:
            rule_map[rule["ID"]] = rule

        rule_diff = data.recursive_diff(before_rule_map, rule_map, ignore_order=True)

        if rule_diff:
            update_payload["Rules"] = rules
            payload["Rules"] = rules
        else:
            payload["Rules"] = before.get("rules")

    if update_payload:
        result["ret"] = {}

        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.s3.put_bucket_replication(
                ctx=ctx,
                Bucket=resource_id,
                ReplicationConfiguration=payload,
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.s3.bucket_replication", name=resource_id
            )
        if "Role" in update_payload:
            result["ret"]["role"] = update_payload["Role"]
        if "Rules" in update_payload:
            result["ret"]["rules"] = update_payload["Rules"]

    return result



import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    rest_api_id: str,
    http_method: str,
    parent_resource_id: str,
    resource_id: str = None,
    api_key_required: bool = False,
    authorization_type: str = None,
) -> Dict[str, Any]:
    r"""
    Creates an API Gateway Method resource

    Args:
        name(string): Idem name of the resource.
        rest_api_id(string): The string identifier of the associated Rest API.
        http_method(string): String that specifies the method request's HTTP method type.
        parent_resource_id(string): AWS Parent Resource ID.
        authorization_type(string, optional): The method's authorization type. Valid values are NONE for open access,
            AWS_IAM for using AWS IAM permissions, CUSTOM for using a custom authorizer, or
            COGNITO_USER_POOLS for using a Cognito user pool. Necessary for creation but not for updating.
        resource_id(string, optional): Idem Resource ID that is generated once the resource is created.
        api_key_required(bool, optional): A boolean flag specifying whether a valid ApiKey is required to invoke this method.

        Parameters not yet supported by Idem:
        authorization_scopes(list, optional): A list of authorization scopes configured on the method. The scopes are
            used with a COGNITO_USER_POOLS authorizer to authorize the method invocation. The authorization works by
            matching the method scopes against the scopes parsed from the access token in the incoming request.
            The method invocation is authorized if any method scopes matches a claimed scope in the access token.
            Otherwise, the invocation is not authorized. When the method scope is configured, the client must provide
            an access token instead of an identity token for authorization purposes.
        authorizer_id(string, optional): The identifier of an Authorizer to use on this method. The authorizationType must be CUSTOM .
        operation_name(string, optional): A human-friendly operation identifier for the method.
        request_parameters(dict, optional): A key-value map defining required or optional method request parameters that
            can be accepted by API Gateway. A key is a method request parameter name matching the pattern of
            method.request.{location}.{name} , where location is querystring , path , or header and name is a valid and
            unique parameter name. The value associated with the key is a Boolean flag indicating whether the parameter
            is required (true ) or optional (false ). The method request parameter names defined here are available in
            Integration to be mapped to integration request parameters or templates.
        request_models(dict, optional): A key-value map specifying data schemas, represented by Model resources,
            (as the mapped value) of the request payloads of given content types (as the mapping key).
        request_validator_id(string, optional): The identifier of a RequestValidator for request validation.



    Request Syntax:
        [idem_test_aws_apigateway_method]:
          aws.apigateway.method.present:
            - name: 'string'
            - rest_api_id: 'string'
            - parent_resource_id: 'string'
            - http_method: 'string'
            - authorization_type: 'string'

    Returns:
        Dict[str, Any]

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.apigateway.method.get(
            ctx,
            name=name,
            resource_id=resource_id,
        )
        if not before["result"] or not before["ret"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result
        result["old_state"] = before["ret"]
        result["new_state"] = copy.deepcopy(result["old_state"])

        update_parameters = {
            "http_method": http_method,
            "authorization_type": authorization_type,
            "api_key_required": api_key_required,
        }

        update_ret = await hub.exec.aws.apigateway.method.update(
            ctx,
            old_state=result["old_state"],
            update_parameters=update_parameters,
        )
        if not update_ret["result"]:
            result["result"] = False
            result["comment"] = update_ret["comment"]
            return result
        result["comment"] = result["comment"] + update_ret["comment"]
        resource_updated = bool(update_ret["ret"])

        if resource_updated and ctx.get("test", False):
            result["new_state"].update(update_ret["ret"])

        if resource_updated:
            if ctx.get("test", False):
                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.apigateway.resource", name=name
                )
                return result
    else:
        resource_id = f"{rest_api_id}-{parent_resource_id}-{http_method}"

        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": resource_id,
                    "rest_api_id": rest_api_id,
                    "parent_resource_id": parent_resource_id,
                    "http_method": http_method,
                    "authorization_type": authorization_type,
                },
            )

            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigateway.method", name=name
            )
            return result

        ret = await hub.exec.boto3.client.apigateway.put_method(
            ctx,
            restApiId=rest_api_id,
            resourceId=parent_resource_id,
            httpMethod=http_method,
            authorizationType=authorization_type,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigateway.method", name=name
        )

    if (not before) or resource_updated:
        after = await hub.exec.aws.apigateway.method.get(
            ctx,
            name=name,
            resource_id=resource_id,
        )

        resource_translated = after["ret"]
        result["new_state"] = resource_translated
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    rest_api_id: str,
    http_method: str,
    parent_resource_id: str,
    resource_id: str = None,
) -> Dict[str, Any]:

    r"""
    Deletes an API Gateway Method resource

    Args:
        name(string): An Idem name of the resource.
        parent_resource_id(string): AWS Parent Resource ID.
        rest_api_id(string): The string identifier of the associated RestApi.
        http_method(string): The HTTP verb of the Method resource.
        resource_id(string, optional): The resource identifier for the Method resource. Idem automatically considers
            this resource being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigateway_method:
              aws.apigateway.method.absent:
                - name: value
                - parent_resource_id: value
                - rest_api_id: value
                - resource_id: value
                - http_method: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        # if resource_id isn't specified, the resource is considered to be absent.
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.method", name=name
        )
        return result

    before_ret = await hub.exec.aws.apigateway.method.get(
        ctx,
        name=name,
        resource_id=resource_id,
    )

    if not before_ret["result"]:
        result["comment"] = before_ret["comment"]
        result["result"] = False
        return result

    if not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.method", name=name
        )

    result["old_state"] = before_ret["ret"]

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigateway.method", name=name
        )
        return result

    delete_ret = await hub.exec.boto3.client.apigateway.delete_method(
        ctx,
        restApiId=rest_api_id,
        resourceId=parent_resource_id,
        httpMethod=http_method,
    )
    if not delete_ret["result"]:
        result["result"] = False
        result["comment"] = delete_ret["comment"]
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.apigateway.method", name=name
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Describe the API Gateway Methods associated with a specific Rest API.

    Returns a list of apigateway.method descriptions

    Returns:
        Dict[str, Any]


    Examples:

        .. code-block:: bash

            $ idem describe aws.apigateway.method

    """

    result = {}

    get_rest_apis_ret = await hub.exec.boto3.client.apigateway.get_rest_apis(ctx)
    if not get_rest_apis_ret["result"]:
        hub.log.debug(f"Could not get Rest Apis {get_rest_apis_ret['comment']}")
        return result

    for rest_api in get_rest_apis_ret["ret"]["items"]:
        rest_api_id = rest_api.get("id")
        get_resources_ret = await hub.exec.boto3.client.apigateway.get_resources(
            ctx, restApiId=rest_api_id
        )
        if not get_resources_ret["result"]:
            hub.log.debug(f"Could not describe resource {get_resources_ret['comment']}")
            return result

        if get_resources_ret["ret"]["items"] is not None:
            for resource in get_resources_ret["ret"]["items"]:
                parent_resource_id = resource.get("id")
                if resource.get("resourceMethods") is not None:
                    for resource_method in resource.get("resourceMethods"):
                        resource_id = (
                            f"{rest_api_id}-{parent_resource_id}-{resource_method}"
                        )

                        ret = await hub.exec.aws.apigateway.method.get(
                            ctx, resource_id=resource_id
                        )

                        resource_translated = ret["ret"]

                        result[resource_translated["resource_id"]] = {
                            "aws.apigateway.method.present": [
                                {parameter_key: parameter_value}
                                for parameter_key, parameter_value in resource_translated.items()
                            ]
                        }

    return result



import datetime
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource", "soft_fail"]

TREQ = {
    "present": {
        "require": [
            "aws.ec2.vpc.present",
            "aws.ec2.subnet.present",
            "aws.iam.role.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    *,
    # From DescribeInstances
    image_id: str = None,
    instance_type: str = None,
    volume_attachments: Dict[str, str] = None,
    ebs_optimized: bool = None,
    kernel_id: str = None,
    subnet_id: str = None,
    network_interfaces: List[
        make_dataclass(
            "InstanceNetworkInterfaceSpecification",
            [
                ("AssociatePublicIpAddress", bool, field(default=None)),
                ("DeleteOnTermination", bool, field(default=None)),
                ("Description", str, field(default=None)),
                ("DeviceIndex", int, field(default=None)),
                ("Groups", List[str], field(default=None)),
                ("Ipv6AddressCount", int, field(default=None)),
                (
                    "Ipv6Addresses",
                    List[
                        make_dataclass(
                            "InstanceIpv6Address",
                            [("Ipv6Address", str, field(default=None))],
                        )
                    ],
                    field(default=None),
                ),
                ("NetworkInterfaceId", str, field(default=None)),
                ("PrivateIpAddress", str, field(default=None)),
                (
                    "PrivateIpAddresses",
                    List[
                        make_dataclass(
                            "PrivateIpAddressSpecification",
                            [
                                ("Primary", bool, field(default=None)),
                                ("PrivateIpAddress", str, field(default=None)),
                            ],
                        )
                    ],
                    field(default=None),
                ),
                ("SecondaryPrivateIpAddressCount", int, field(default=None)),
                ("SubnetId", str, field(default=None)),
                ("AssociateCarrierIpAddress", bool, field(default=None)),
                ("InterfaceType", str, field(default=None)),
                ("NetworkCardIndex", int, field(default=None)),
                (
                    "Ipv4Prefixes",
                    List[
                        make_dataclass(
                            "Ipv4Prefix", [("Ipv4Prefix", str, field(default=None))]
                        )
                    ],
                    field(default=None),
                ),
                ("Ipv4PrefixCount", int, field(default=None)),
                (
                    "Ipv6Prefixes",
                    List[
                        make_dataclass(
                            "Ipv6Prefix", [("Ipv6Prefix", str, field(default=None))]
                        )
                    ],
                    field(default=None),
                ),
                ("Ipv6PrefixCount", int, field(default=None)),
            ],
        )
    ] = None,
    monitoring_enabled: bool = None,
    source_dest_check: bool = None,
    running: bool = None,
    private_ip_address: str = None,
    owner_id: str = None,
    # Placement Options
    availability_zone: str = None,
    affinity: str = None,
    group_name: str = None,
    partition_number: int = None,
    host_id: str = None,
    tenancy: str = None,
    spread_domain: str = None,
    host_resource_group_arn: str = None,
    # From launchTemplate
    user_data: str = None,
    disable_api_termination: bool = None,
    ram_disk_id: str = None,
    tags: Dict[str, str] = None,
    iam_profile_arn: str = None,
    instance_initiated_shutdown_behavior: str = None,
    elastic_inference_accelerators: List[
        make_dataclass(
            "ElasticInferenceAccelerator",
            [("Type", str), ("Count", int, field(default=None))],
        )
    ] = None,
    elastic_gpu_specifications: List[
        make_dataclass("ElasticGpuSpecification", [("Type", str, field(default=None))])
    ] = None,
    auto_recovery_enabled: bool = None,
    sriov_net_support: str = None,
    key_name: str = None,
    # Enclave options
    nitro_enclave_enabled: bool = None,
    # Can only be changed on initial creation of an instance as far as we know so far
    capacity_reservation_specification: make_dataclass(
        "CapacityReservationSpecification",
        [
            ("CapacityReservationPreference", str, field(default=None)),
            (
                "CapacityReservationTarget",
                make_dataclass(
                    "CapacityReservationTarget",
                    [
                        ("CapacityReservationId", str, field(default=None)),
                        (
                            "CapacityReservationResourceGroupArn",
                            str,
                            field(default=None),
                        ),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    client_token: str = None,
    root_device_name: str = None,
    product_codes: List[Dict[str, str]] = None,
    reservation_id: str = None,
    block_device_mappings: List[
        make_dataclass(
            "BlockDeviceMapping",
            [
                ("DeviceName", str, field(default=None)),
                ("VirtualName", str, field(default=None)),
                ("Ebs", Dict, field(default=None)),
                ("NoDevice", str, field(default=None)),
            ],
        )
    ] = None,
    license_arns: List[str] = None,
    hibernation_enabled: bool = None,
    # Instance Market Options
    market_type: str = None,
    # Spot Options
    max_price: str = None,
    spot_instance_type: str = None,
    block_duration_minutes: int = None,
    valid_until: str = None,
    instance_interruption_behavior: str = None,
    # Credit Specification
    cpu_credits: str = None,
    cpu_core_count: int = None,
    cpu_threads_per_core: int = None,
    # Metadata options
    http_tokens: str = None,
    http_put_response_hop_limit: int = None,
    http_endpoint_enabled: bool = None,
    http_protocol_ipv6_enabled: bool = None,
    metadata_tags_enabled: bool = None,
    # Private DNS Name options
    hostname_type: str = None,
    enable_resource_name_dns_a_record: bool = None,
    enable_resource_name_dns_aaaa_record: bool = None,
    # Capacity Reservation Options
    capacity_reservation_preference: str = None,
    instance_requirements: Dict[str, Any] = None,
    # idem-heist options
    bootstrap: List[Dict[str, str]] = None,
    **kwargs,
) -> Dict[str, Any]:
    """Launches an instance using an AMI for which you have permissions.

    You can specify a number of options, or leave the default options.

    The following rules apply:

        - [EC2-VPC] If you don't specify a subnet ID, we choose a default subnet from your default VPC for you.
            If you don't have a default VPC, you must specify a subnet ID in the request.
        - [EC2-Classic] If don't specify an Availability Zone, we choose one for you.
            Some instance types must be launched into a VPC.
            If you do not have a default VPC, or if you do not specify a subnet ID, the request fails.
            For more information, see Instance types available only in a VPC.
        - [EC2-VPC] All instances have a network interface with a primary private IPv4 address.
            If you don't specify this address, we choose one from the IPv4 range of your subnet.
            Not all instance types support IPv6 addresses.
            For more information, see Instance types.
        - If you don't specify a security group ID, we use the default security group.
            For more information, see Security groups.
        - If any of the AMIs have a product code attached for which the user has not subscribed, the request fails.
        - You can create a launch template, which is a resource that contains the parameters to launch an instance.
            You can specify the launch template instead of specifying the launch parameters.
            An instance is ready for you to use when it's in the running state.
        - You can tag instances and EBS volumes during launch, after launch, or both.
        - Linux instances have access to the public key of the key pair at boot.
            You can use this key to provide secure access to the instance.
            Amazon EC2 public images use this feature to provide secure access without passwords.
            For more information, see Key pairs.

    Args:
        hub:

        ctx:

        name(str):
            An Idem name of the resource.

        resource_id(str):
            AWS Ec2 Instance ID.

        image_id(str):
            The ID of an AMI.

        instance_type(str):
            The instance type to use for this instance on creation.

        volume_attachments(dict[str, str]):
            The instance device name mapped to a volume id.

        ebs_optimized(bool):
            Indicates whether the instance is optimized ofr Amazon EBS I/O.

        kernel_id(str):
            The kernel associated with this instance, if applicable.

        subnet_id(str):
            The ID of the subnet in which the instance is running.

        network_interfaces(list[dict[str, Any]], Optional):
            The network interfaces to associate with the instance. If you specify a network interface, you must specify
            any security groups and subnets as part of the network interface. Defaults to None.

            * AssociatePublicIpAddress (bool, Optional):
                Indicates whether to assign a public IPv4 address to an instance you launch in a VPC. The public
                IP address can only be assigned to a network interface for eth0, and can only be assigned to a
                new network interface, not an existing one. You cannot specify more than one network interface
                in the request. If launching into a default subnet, the default value is true.

            * DeleteOnTermination (bool, Optional):
                If set to true, the interface is deleted when the instance is terminated. You can specify true only if
                creating a new network interface when launching an instance.

            * Description (str, Optional):
                The description of the network interface. Applies only if creating a network interface when launching
                an instance.

            * DeviceIndex (int, Optional):
                The position of the network interface in the attachment order. A primary network interface has a
                device index of 0. If you specify a network interface when launching an instance, you must
                specify the device index.

            * Groups (list[str], Optional):
                The IDs of the security groups for the network interface. Applies only if creating a network
                interface when launching an instance.

            * Ipv6AddressCount (int, Optional):
                A number of IPv6 addresses to assign to the network interface. Amazon EC2 chooses the IPv6
                addresses from the range of the subnet. You cannot specify this option and the option to assign
                specific IPv6 addresses in the same request. You can specify this option if you've specified a
                minimum number of instances to launch.

            * Ipv6Addresses (list[dict[str, Any]], Optional):
                One or more IPv6 addresses to assign to the network interface. You cannot specify this option
                and the option to assign a number of IPv6 addresses in the same request. You cannot specify this
                option if you've specified a minimum number of instances to launch.

                * Ipv6Address (str, Optional):
                    The IPv6 address.

            * NetworkInterfaceId (str, Optional):
                The ID of the network interface. If you are creating a Spot Fleet, omit this parameter because
                you can’t specify a network interface ID in a launch specification.

            * PrivateIpAddress (str, Optional):
                The private IPv4 address of the network interface. Applies only if creating a network interface
                when launching an instance. You cannot specify this option if you're launching more than one
                instance in a RunInstances request.

            * PrivateIpAddresses (list[dict[str, Any]], Optional):
                One or more private IPv4 addresses to assign to the network interface. Only one private IPv4
                address can be designated as primary. You cannot specify this option if you're launching more
                than one instance in a RunInstances request.

                * Primary (bool, Optional):
                    Indicates whether the private IPv4 address is the primary private IPv4 address. Only one IPv4
                    address can be designated as primary.

                * PrivateIpAddress (str, Optional):
                    The private IPv4 addresses.

            * SecondaryPrivateIpAddressCount (int, Optional):
                The number of secondary private IPv4 addresses. You can't specify this option and specify more
                than one private IP address using the private IP addresses option. You cannot specify this
                option if you're launching more than one instance in a RunInstances request.

            * SubnetId (str, Optional):
                The ID of the subnet associated with the network interface. Applies only if creating a network
                interface when launching an instance.

            * AssociateCarrierIpAddress (bool, Optional):
                Indicates whether to assign a carrier IP address to the network interface. You can only assign a
                carrier IP address to a network interface that is in a subnet in a Wavelength Zone. For more
                information about carrier IP addresses, see Carrier IP addresses in the Amazon Web Services
                Wavelength Developer Guide.

            * InterfaceType (str, Optional):
                The type of network interface. Valid values: interface | efa.

            * NetworkCardIndex (int, Optional):
                The index of the network card. Some instance types support multiple network cards. The primary
                network interface must be assigned to network card index 0. The default is network card index 0.
                If you are using RequestSpotInstances to create Spot Instances, omit this parameter because you
                can’t specify the network card index when using this API. To specify the network card index, use
                RunInstances.

            * Ipv4Prefixes (list[dict[str, Any]], Optional):
                One or more IPv4 delegated prefixes to be assigned to the network interface. You cannot use this
                option if you use the Ipv4PrefixCount option.

                * Ipv4Prefix (str, Optional):
                    The IPv4 prefix. For information, see  Assigning prefixes to Amazon EC2 network interfaces in
                    the Amazon Elastic Compute Cloud User Guide.

            * Ipv4PrefixCount (int, Optional):
                The number of IPv4 delegated prefixes to be automatically assigned to the network interface. You
                cannot use this option if you use the Ipv4Prefix option.

            * Ipv6Prefixes (list[dict[str, Any]], Optional):
                One or more IPv6 delegated prefixes to be assigned to the network interface. You cannot use this
                option if you use the Ipv6PrefixCount option.

                * Ipv6Prefix (str, Optional):
                    The IPv6 prefix.

            * Ipv6PrefixCount (int, Optional):
                The number of IPv6 delegated prefixes to be automatically assigned to the network interface. You
                cannot use this option if you use the Ipv6Prefix option.

        monitoring_enabled(bool):
            Indicates whether detailed monitoring is enabled.

        source_dest_check(bool):
            Indicates whether source/destination checking is enabled.

        running(bool):
            Indicates whether the instance should be in the "running" state.

        private_ip_address(str):
            The Ipv4 address of the network interface within the subnet.

        owner_id(str):
            The ID of the AWS account that owns the reservation.

        availability_zone(str):
            The Availability Zone of the instance.

        affinity(str):
            The affinity setting for the instance on the Dedicated Host.

        group_name(str):
            The affinity setting for the instance on the Dedicated Host.

        partition_number(int):
            The number of the partition that the instance is in.

        host_id(str):
            The ID of the Dedicated Host on which the instance resides.

        tenancy(str):
            The tenancy of the instance (if the instance is running in a VPC). An instance with a tenancy of dedicated runs on single-tenant hardware.

        spread_domain(str):
            Not yet documented by AWS.

        host_resource_group_arn(str):
            The ARN of the host resource group in which to launch the instances.

        user_data(str):
            The user data for the instance.

        disable_api_termination(bool):
            Indicates that an instance cannot be terminated using the Amazon Ec2 console, command line tool, or API.

        ram_disk_id(str):
            The ID of the RAM disk, if applicable.

        tags(dict, Optional):
            The tags to apply to the resource. Defaults to None.

        iam_profile_arn(str):
            The IAM instance profile ARN.

        instance_initiated_shutdown_behavior(str): Indicates whether an instance stops or terminates when you initiate
            shutdown from the instance (using the operating system command for system shutdown).

        elastic_inference_accelerators(list[dict[str, Any]], Optional):
            An elastic inference accelerator to associate with the instance. Elastic inference accelerators
            are a resource you can attach to your Amazon EC2 instances to accelerate your Deep Learning (DL)
            inference workloads. You cannot specify accelerators from different generations in the same
            request. Defaults to None.

            * Type (str):
                The type of elastic inference accelerator. The possible values are eia1.medium, eia1.large,
                eia1.xlarge, eia2.medium, eia2.large, and eia2.xlarge.

            * Count (int, Optional):
                The number of elastic inference accelerators to attach to the instance.  Default: 1

        elastic_gpu_specifications(list[dict[str, Any]], Optional):
            An elastic GPU to associate with the instance. An Elastic GPU is a GPU resource that you can
            attach to your Windows instance to accelerate the graphics performance of your applications. For
            more information, see Amazon EC2 Elastic GPUs in the Amazon EC2 User Guide. Defaults to None.

            * Type (str): The type of Elastic Graphics accelerator. For more information about the values to specify for
                Type, see Elastic Graphics Basics, specifically the Elastic Graphics accelerator column, in the
                Amazon Elastic Compute Cloud User Guide for Windows Instances.

        auto_recovery_enabled(bool):
            Disables the automatic recovery behavior of your instance or sets it to default.

        sriov_net_support(str):
            Specifies whether enhanced networking with the Intel 82599 Virtual Function interface is enabled.

        key_name(str):
            The name of the keypair.

        nitro_enclave_enabled(bool):
            Indicates whether the instance is enabled for AWS Nitro Enclaves.

        capacity_reservation_specification(dict[str, Any], Optional):
            Information about the Capacity Reservation targeting option. If you do not specify this
            parameter, the instance's Capacity Reservation preference defaults to open, which enables it to
            run in any open Capacity Reservation that has matching attributes (instance type, platform,
            Availability Zone). Defaults to None.

            * CapacityReservationPreference (str, Optional):
                Indicates the instance's Capacity Reservation preferences.
                Possible preferences include:

                * open:
                    The instance can run in any open Capacity Reservation that has matching attributes (instance
                    type, platform, Availability Zone).

                * none:
                    The instance avoids running in a Capacity Reservation even if one is available. The instance runs as an On-Demand Instance.

            * CapacityReservationTarget (dict[str, Any], Optional):
                Information about the target Capacity Reservation or Capacity Reservation group.

                * CapacityReservationId (str, Optional):
                    The ID of the Capacity Reservation in which to run the instance.

                * CapacityReservationResourceGroupArn (str, Optional):
                    The ARN of the Capacity Reservation resource group in which to run the instance.

        client_token(str):
            The idempotency token for the instance.

        root_device_name(str):
            The device name of the root device (for example, /dev/sda1).

        product_codes(list[dict[str, str]]:
            The product codes attached to the instance, if applicable.

        reservation_id(str):
            The ID of the reservation

        block_device_mappings(list[dict[str, Any]], Optional):
            The block device mapping, which defines the EBS volumes and instance store volumes to attach to
            the instance at launch. For more information, see Block device mappings in the Amazon EC2 User
            Guide. Defaults to None.

            * DeviceName (str, Optional):
                The device name (for example, /dev/sdh or xvdh).

            * VirtualName (str, Optional):
                The virtual device name (ephemeralN). Instance store volumes are numbered starting from 0. An
                instance type with 2 available instance store volumes can specify mappings for ephemeral0 and
                ephemeral1. The number of available instance store volumes depends on the instance type. After
                you connect to the instance, you must mount the volume. NVMe instance store volumes are
                automatically enumerated and assigned a device name. Including them in your block device mapping
                has no effect. Constraints: For M3 instances, you must specify instance store volumes in the
                block device mapping for the instance. When you launch an M3 instance, we ignore any instance
                store volumes specified in the block device mapping for the AMI.

            * Ebs (dict[str, Any], Optional):
                Parameters used to automatically set up EBS volumes when the instance is launched.

                * DeleteOnTermination (bool, Optional):
                    Indicates whether the EBS volume is deleted on instance termination. For more information, see
                    Preserving Amazon EBS volumes on instance termination in the Amazon EC2 User Guide.

                * Iops (int, Optional): The number of I/O operations per second (IOPS). For gp3, io1, and io2 volumes,
                    this represents the number of IOPS that are provisioned for the volume. For gp2 volumes, this
                    represents the baseline performance of the volume and the rate at which the volume accumulates I/O
                    credits for bursting. The following are the supported values for each volume type:
                    gp3: 3,000-16,000 IOPS    io1: 100-64,000 IOPS    io2: 100-64,000 IOPS   For io1 and io2 volumes,
                    we guarantee 64,000 IOPS only for Instances built on the Nitro System. Other instance families
                    guarantee performance up to 32,000 IOPS. This parameter is required for io1 and io2 volumes.
                    The default for gp3 volumes is 3,000 IOPS. This parameter is not supported for gp2, st1, sc1, or
                    standard volumes.

                * SnapshotId (str, Optional):
                    The ID of the snapshot.

                * VolumeSize (int, Optional):
                    The size of the volume, in GiBs. You must specify either a snapshot ID or a volume size. If you
                    specify a snapshot, the default is the snapshot size. You can specify a volume size that is
                    equal to or larger than the snapshot size. The following are the supported volumes sizes for
                    each volume type:    gp2 and gp3:1-16,384    io1 and io2: 4-16,384    st1 and sc1: 125-16,384
                    standard: 1-1,024.

                * VolumeType (str, Optional):
                    The volume type. For more information, see Amazon EBS volume types in the Amazon EC2 User Guide.
                    If the volume type is io1 or io2, you must specify the IOPS that the volume supports.

                * KmsKeyId (str, Optional):
                    Identifier (key ID, key alias, ID ARN, or alias ARN) for a customer managed CMK under which the
                    EBS volume is encrypted. This parameter is only supported on BlockDeviceMapping objects called
                    by RunInstances, RequestSpotFleet, and RequestSpotInstances.

                * Throughput (int, Optional):
                    The throughput that the volume supports, in MiB/s. This parameter is valid only for gp3 volumes.
                    Valid Range: Minimum value of 125. Maximum value of 1000.

                * OutpostArn (str, Optional):
                    The ARN of the Outpost on which the snapshot is stored. This parameter is only supported on
                    BlockDeviceMapping objects called by  CreateImage.

                * Encrypted (bool, Optional):
                    Indicates whether the encryption state of an EBS volume is changed while being restored from a
                    backing snapshot. The effect of setting the encryption state to true depends on the volume
                    origin (new or from a snapshot), starting encryption state, ownership, and whether encryption by
                    default is enabled. For more information, see Amazon EBS encryption in the Amazon EC2 User
                    Guide. In no case can you remove encryption from an encrypted volume. Encrypted volumes can only
                    be attached to instances that support Amazon EBS encryption. For more information, see Supported
                    instance types. This parameter is not returned by DescribeImageAttribute.

            * NoDevice (str, Optional): To omit the device from the block device mapping, specify an empty string. When this property is
                specified, the device is removed from the block device mapping regardless of the assigned value.

        license_arns(list[str]):
            The license configuration arns.

        hibernation_enabled(bool):
            Indicates whether the instance is configured for hibernation.

        market_type(str):
            The market (purchasing) option for the instance.

        max_price(str):
            The maximum hourly price you're willing to pay for the Spot Instances.

        spot_instance_type(str):
            The Spot Instance request type. Persistent Spot Instance requests are only supported when the instance
            interruption behavior is either hibernate or stop.

        block_duration_minutes(int):
            Deprecated.

        valid_until(str):
            The end date of the request, in UTC format (YYYY -MM -DD T*HH* :MM :SS Z). Supported only for persistent requests.

        instance_interruption_behavior(str):
            The behavior when a Spot Instance is interrupted. The default is terminate.

        cpu_credits(str):
            The credit option for CPU usage of a T2, T3, or T3a instance. Valid values are standard and unlimited.

        cpu_core_count(int):
            The number of CPU cores for the instance.

        cpu_threads_per_core(int):
            The number of threads per CPU core. To disable multithreading for the instance, specify a value of1 .
            Otherwise, specify the default value of 2.

        http_tokens(str):
            The state of token usage for your instance metadata requests. If the state is optional, you can
            choose to retrieve instance metadata with or without a signed token header on your request. If
            you retrieve the IAM role credentials without a token, the version 1.0 role credentials are
            returned. If you retrieve the IAM role credentials using a valid signed token, the version 2.0
            role credentials are returned. If the state is required, you must send a signed token header
            with any instance metadata retrieval requests. In this state, retrieving the IAM role
            credentials always returns the version 2.0 credentials; the version 1.0 credentials are not
            available. Default: Optional.

        http_put_response_hop_limit(int):
            The desired HTTP PUT response hop limit for instance metadata requests. The larger the number,
            the further instance metadata requests can travel. Default: 1 Possible values: Integers from 1
            to 64.

        http_endpoint_enabled(bool):
            Enables or disables the HTTP metadata endpoint on your instances. If you specify a value of
            disabled, you cannot access your instance metadata. Default: enabled.

        http_protocol_ipv6_enabled(str):
            Enables or disables the IPv6 endpoint for the instance metadata service.

        metadata_tags_enabled(bool):
            Set to enabled to allow access to instance tags from the instance metadata. Set to disabled to
            turn off access to instance tags from the instance metadata. For more information, see Work with
            instance tags using the instance metadata. Default: disabled.

        hostname_type(str):
            The type of hostname for EC2 instances.
            For IPv4 only subnets, an instance DNS name must be based on the instance IPv4 address.
            For IPv6 only subnets, an instance DNS name must be based on the instance ID.
            For dual-stack subnets, you can specify whether DNS names use the instance IPv4 address or the instance ID.

        enable_resource_name_dns_a_record(bool):
            Indicates whether to respond to DNS queries for instance hostnames with DNS A records.

        enable_resource_name_dns_aaaa_record(bool):
            Indicates whether to respond to DNS queries for instance hostnames with DNS A records.

        capacity_reservation_preference(str):
            Indicates the instance's Capacity Reservation preferences.

        instance_requirements(dict[str, Any]):
            The attributes for the instance type. When you specify instance attributes, Amazon EC2 will identify
            instance types with these attributes.

        bootstrap(list[dict[str, Any]]):
            Bootstr options for provisioning an instance with "heist".

    Request Syntax:
       .. code-block:: sls

          [instance-resource-id]:
            aws.ec2.instance.present:
              - name: "string"
              - image_id: "string"
              - tags:
                  - Key: "string"
                    Value: "string"
              - bootstrap:
                    - heist_manager: "string"
                      artifact_version: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.ec2.instance.present:
                - name: value
                - image_id: my_ami_id-0000000000000000
                - tags:
                    my_tag_key_1: my_tag_value_1
                    my_tag_key_2: my_tag_value_2
                - bootstrap:
                    - heist_manager: salt.minion
                      artifact_version: 1234

    """
    # Get all the parameters passed to this function as a single dictionary
    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "name", "resource_id", "kwargs") and v is not None
    }

    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)

    for key in kwargs:
        result["comment"] += [f"Not explicitly supported keyword argument: '{key}'"]

    # Get the resource_id from ESM
    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    if resource_id:
        # Assume that the instance already exists since we have a resource_id
        resource = hub.tool.boto3.resource.create(ctx, "ec2", "Instance", resource_id)
        get = await hub.exec.aws.ec2.instance.get(
            ctx, name=name, resource_id=resource_id
        )
        current_state = result["old_state"] = get.ret

        if not result["old_state"]:
            result["comment"] += [
                f"Could not find instance for '{name}' with existing id '{resource_id}'"
            ]
            return result

        result["comment"] += [f"Instance '{name}' already exists"]
    else:
        # Create the instance if it doesn't already exist
        if ctx.test:
            result["new_state"] = hub.tool.aws.ec2.instance.state.test(**desired_state)
            result["comment"] += [f"Would create aws.ec2.instance '{name}'"]
            return result

        # Prepare options that are grouped in the request

        # Prepare instance market options
        spot_options = {}
        if max_price:
            spot_options["MaxPrice"] = max_price
        if spot_instance_type:
            spot_options["SpotInstanceType"] = spot_instance_type
        if block_duration_minutes:
            spot_options["BlockDurationMinutes"] = block_duration_minutes
        if valid_until:
            spot_options["ValidUntil"] = datetime.datetime(valid_until)
        if instance_initiated_shutdown_behavior:
            spot_options[
                "InstanceInterruptionBehavior"
            ] = instance_interruption_behavior
        if market_type:
            instance_market_options = {
                "MarketType": market_type,
                "SpotOptions": spot_options,
            }
        elif spot_options:
            instance_market_options = {"SpotOptions": spot_options}
        else:
            instance_market_options = None

        # Prepare Metadata options
        metadata_options = {}
        if http_tokens:
            metadata_options["HttpTokens"] = http_tokens
        if http_put_response_hop_limit:
            metadata_options["HttpPutResponseHopLimit"] = http_put_response_hop_limit
        if http_endpoint_enabled is not None:
            metadata_options["HttpEndpoint"] = (
                "enabled" if http_endpoint_enabled else "disabled"
            )
        if http_protocol_ipv6_enabled is not None:
            metadata_options["HttpProtocolIpv6"] = (
                "enabled" if http_protocol_ipv6_enabled else "disabled"
            )
        if metadata_tags_enabled is not None:
            metadata_options["InstanceMetadataTags"] = (
                "enabled" if metadata_tags_enabled else "disabled"
            )

        # Prepare Private DNS Name Options
        private_dns_name_options = {}
        if hostname_type:
            private_dns_name_options["HostnameType"] = hostname_type
        if enable_resource_name_dns_a_record is not None:
            private_dns_name_options[
                "EnableResourceNameDnsARecord"
            ] = enable_resource_name_dns_a_record
        if enable_resource_name_dns_aaaa_record is not None:
            private_dns_name_options[
                "EnableResourceNameDnsAAAARecord"
            ] = enable_resource_name_dns_aaaa_record

        # Prepare Cpu Options
        cpu_options = {}
        if cpu_core_count:
            cpu_options["CoreCount"] = cpu_core_count
        if cpu_threads_per_core:
            cpu_options["ThreadsPerCore"] = cpu_threads_per_core

        # Prepare Placement Options
        placement = {}
        if availability_zone:
            placement["AvailabilityZone"] = availability_zone
        if affinity:
            placement["Affinity"] = affinity
        if group_name:
            placement["GroupName"] = group_name
        if partition_number:
            placement["PartitionNumber"] = partition_number
        if host_id:
            placement["HostId"] = host_id
        if tenancy:
            placement["Tenancy"] = tenancy
        if spread_domain:
            placement["SpreadDomain"] = spread_domain
        if host_resource_group_arn:
            placement["HostResourceGroupArn"] = host_resource_group_arn

        # Prepare TagSpecifications
        if tags:
            tag_specifications = [
                {
                    "ResourceType": "instance",
                    "Tags": [{"Key": k, "Value": v} for k, v in tags.items()],
                }
            ]
        else:
            tag_specifications = None

        # Prepare Maintenance Options
        maintenance_options = {}
        if auto_recovery_enabled is not None:
            maintenance_options["AutoRecovery"] = (
                "disabled" if auto_recovery_enabled is False else "default"
            )

        # Prepare Hibernation Options
        hibernation_options = {}
        if hibernation_enabled is not None:
            hibernation_options["Configured"] = hibernation_enabled

        # Prepare Credit Specifications
        credit_specifications = {}
        if cpu_credits is not None:
            credit_specifications["CpuCredits"] = cpu_credits

        # Prepare Enclave Options
        enclave_options = {}
        if nitro_enclave_enabled:
            enclave_options["Enabled"] = nitro_enclave_enabled

        # Prepare Monitoring Options
        monitoring_options = {}
        if monitoring_enabled:
            monitoring_options["Enabled"] = monitoring_enabled

        # Prepare Capacity Reservation Specification
        capacity_reservation_specification = {}
        if capacity_reservation_preference:
            capacity_reservation_specification[
                "CapacityReservationPreference"
            ] = capacity_reservation_preference

        # Create a brand new instance with minimal arguments, it will be updated after creation
        create_ret = await hub.exec.boto3.client.ec2.run_instances(
            ctx,
            ClientToken=client_token,
            MaxCount=1,
            MinCount=1,
            ImageId=image_id,
            InstanceType=instance_type,
            BlockDeviceMappings=block_device_mappings,
            EbsOptimized=ebs_optimized,
            KernelId=kernel_id,
            Placement=placement or None,
            SubnetId=subnet_id,
            NetworkInterfaces=network_interfaces,
            Monitoring=monitoring_options or None,
            PrivateIpAddress=private_ip_address,
            UserData=user_data,
            DisableApiTermination=disable_api_termination,
            InstanceInitiatedShutdownBehavior=instance_initiated_shutdown_behavior,
            EnclaveOptions=enclave_options or None,
            RamDiskId=ram_disk_id,
            TagSpecifications=tag_specifications,
            ElasticGpuSpecifications=elastic_gpu_specifications,
            ElasticInferenceAccelerators=elastic_inference_accelerators,
            IamInstanceProfile={"Arn": iam_profile_arn} if iam_profile_arn else None,
            KeyName=key_name,
            InstanceMarketOptions=instance_market_options,
            CreditSpecification=credit_specifications or None,
            CpuOptions=cpu_options or None,
            CapacityReservationSpecification=capacity_reservation_specification or None,
            LicenseSpecifications=[
                {"LicenseConfigurationArn": license_arn}
                for license_arn in license_arns or []
            ],
            HibernationOptions=hibernation_options or None,
            MetadataOptions=metadata_options or None,
            PrivateDnsNameOptions=private_dns_name_options or None,
            MaintenanceOptions=maintenance_options or None,
        )

        result["result"] &= create_ret.result
        if not create_ret:
            result["comment"] += [create_ret.comment]
            return result

        result["comment"] += [f"Created '{name}'"]
        resource_id = create_ret.ret["Instances"][0]["InstanceId"]
        # This makes sure the created VPC is saved to esm regardless if the subsequent update call fails or not.
        result["new_state"] = dict(name=name, resource_id=resource_id)
        result["force_save"] = True

        present_ret = await hub.tool.aws.ec2.instance.state.convert_to_present(
            ctx, {"Reservations": [create_ret.ret]}
        )
        current_state = present_ret[resource_id]

    if not current_state:
        result["comment"] += [
            f"Unable to get the current_state, instance may still be undergoing creation: '{name}'"
        ]
        result["result"] = False
        return result

    # If no attributes need modification, then return right away
    if all(
        current_state.get(attribute) == new_value
        for attribute, new_value in desired_state.items()
        if new_value is not None
    ):
        result["new_state"] = current_state
        return result

    # Modify all the attributes that need modification
    if ctx.test:
        result["new_state"] = hub.tool.aws.ec2.instance.state.test(**desired_state)
    else:
        resource = hub.tool.boto3.resource.create(ctx, "ec2", "Instance", resource_id)
        hub.log.debug(f"Waiting for instance '{resource.id}' to be created")
        try:
            await hub.tool.boto3.resource.exec(resource, "wait_until_exists")
        except:
            ...
        result["result"] &= await hub.tool.aws.ec2.instance.update.init.apply(
            ctx,
            resource,
            old_value=current_state,
            new_value=desired_state,
            comments=result["comment"],
        )

        get = await hub.exec.aws.ec2.instance.get(ctx, resource_id=resource_id)
        result["new_state"] = get.ret

    # Report successful bootstrapping methods
    # TODO remove this once bootstrapping can be detected with heist
    result["new_state"]["bootstrap"] = [
        bootstrap_options["heist_manager"]
        for bootstrap_options in desired_state.get("boostrap", ())
    ]
    return result


async def absent(
    hub, ctx, name: str, resource_id: str = None, **kwargs
) -> Dict[str, Any]:
    """Shuts down the specified instance.

    Terminated instances remain visible after termination (for approximately one hour).

    Args:
        name(str): The name of the state.
        resource_id(str): An instance id.

    Request Syntax:
        .. code-block:: sls

            [instance-resource-id]:
              aws.ec2.instance.absent:
                - name: "string"
                - resource_id: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.ec2.instance.absent:
                - name: value
    """
    result = dict(
        comment=[], old_state=ctx.old_state, new_state=None, name=name, result=True
    )

    for key in kwargs:
        result["comment"] += [f"Not explicitly supported keyword argument: '{key}'"]

    # Get the resource_id from ESM
    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    # If there still is no resource_id, the instance is gone
    if not resource_id:
        result["comment"] += [f"'{name}' already terminated"]
        return result

    if ctx.test:
        result["comment"] += [f"Would terminate aws.ec2.instance '{name}'"]
        return result

    ret = await hub.exec.boto3.client.ec2.terminate_instances(
        ctx, InstanceIds=[resource_id]
    )
    result["result"] &= ret["result"]
    if not result["result"]:
        result["comment"].append(ret["comment"])
        return result
    result["comment"] += [
        f"Terminated instance '{name}', it will still be visible for about 60 minutes"
    ]

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.ec2.instance
    """
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_instances(ctx)

    if not ret:
        hub.log.debug(f"Could not describe Instances: {ret.comment}")
        return {}

    instances = await hub.tool.aws.ec2.instance.state.convert_to_present(ctx, ret.ret)

    for instance_id, present_state in instances.items():
        result[instance_id] = {
            "aws.ec2.instance.present": [{k: v} for k, v in present_state.items()]
        }

    return result



from typing import Any
from typing import Dict
from typing import List


async def apply(
    hub,
    ctx,
    resource,
    *,
    old_value: Dict[str, Any],
    new_value: Dict[str, Any],
    comments: List[str],
) -> bool:
    """
    Compare the kwargs of the present function to the presentized attributes of the instance
    """
    # TODO day2 operations are still WIP
    return True

    current_state = old_value
    desired_state = new_value

    # Get the current_state/desired state for running
    update_start_state = current_state.get("running")
    update_current_run_state = current_state.get("running")

    update_needs_stopped = []
    update_needs_running = []
    update_group: Dict[str, List[str]] = {}
    update_remainder = []

    # Group attributes together to determine how to do the update
    for attribute, new_attr_value in desired_state.items():
        if desired_state[attribute] is None:
            # There is no new value for the attribute
            continue
        if desired_state[attribute] == current_state.get(attribute):
            # This specific attribute is in the proper state
            continue

        mod = hub.tool.aws.ec2.instance.update[attribute]
        if getattr(mod, "__update_state__", None) == "stopped":
            # This attribute can only be modified while stopped
            update_needs_stopped.append(mod)
        elif getattr(mod, "__update_state__", None) == "running":
            # This attribute can only be modified while running
            update_needs_running.append(mod)
        elif getattr(mod, "__update_group__", None):
            # Some attributes need to be modified together, send their data to a single mod
            group = getattr(mod, "__update_group__")
            if group not in update_group:
                update_group[group] = []
            update_group[group].append(mod)
        else:
            # This attribute has no extra needs
            update_remainder.append(mod)

    # Start with the attributes that work with the instance in the current running state
    # This will minimize number of shutdown/restarts as we update the instance
    if resource.state.name == "running" and update_needs_running:
        ...
        # first_update_group = update_needs_running
        # second_update_group = update_needs_stopped

    try:
        if update_needs_stopped and update_current_run_state is True:
            await hub.exec.aws.ec2.instance.stop(ctx, instance_id="TODO")

            update_current_run_state = False

        # TODO Modify all the attributes that require the instance to be stopped first, stop if needed
        # TODO Group attributes that need to be modified together (like for placement)
        # TODO run all other attribute modifications

        if update_needs_running and current_state.get("running") is False:
            await hub.exec.aws.ec2.instance.start(ctx, instance_id="TODO")
            update_current_run_state = True

        # TODO Modify all the attributes that require the instance to be running first, start if needed
        return True
    finally:
        # Put the state back in the desired/previous state for running
        desired_run_state = desired_state.get("running", update_start_state)
        if desired_run_state != update_current_run_state:
            if desired_run_state is True:
                await hub.exec.aws.ec2.instance.start(ctx, instance_id="TODO")
            else:
                await hub.exec.aws.ec2.instance.stop(ctx, instance_id="TODO")


async def old_update(
    hub, ctx, name, resource, current_state, desired_state, comments, result
):
    # TODO delete this function when the main apply function is done
    for attribute, new_value in desired_state.items():
        if new_value is None:
            # No value has been explicitly given, leave this parameter alone
            continue

        # TODO desired state from ESM is showing changes when it shouldn't...
        old_value = current_state.get(attribute)

        if old_value != new_value:
            # There is a single file dedicated to updating each attribute of an ec2 instance.
            # Organization is key for managing such a large resource.
            # This "present" function should remain mostly the same, don't bloat it!
            # Add an attribute-specific file in idem_aws/tool/aws/ec2/instance/update to manage specific parameters.
            # Be sure to follow the contracts in idem_aws/tool/aws/ec2/instance/update/contracts
            if attribute not in hub.tool.aws.ec2.instance.update._loaded:
                comments += [
                    f"Modifying aws.ec2.instance attribute '{attribute}' is not yet supported"
                ]
                continue

            if ctx.test:
                comments += [f"Would update aws.ec2.instance '{name}': {attribute}"]
                continue

            # Call the appropriate tool to update each parameter that needs updating
            result["result"] &= await hub.tool.aws.ec2.instance.update[attribute].apply(
                ctx,
                resource,
                old_value=old_value,
                new_value=new_value,
                # This list is stored in memory
                # modifying this value in "update.apply" functions will update it in the "result" dictionary
                comments=result["comment"],
            )
            if not result["result"]:
                result["comment"] += [
                    f"Unable to update aws.ec2.instance attribute: {attribute}"
                ]
                break



"""Exec module for managing Lambda provisioned concurrency config."""
__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    function_name: str = None,
    qualifier: str = None,
):
    """Get the provisioned concurrency configuration for a function's alias or version.

    Args:
        name(str): An Idem name of the resource.
        resource_id(str, Optional): The Amazon Resource Name (ARN) of the function. If not supplied, the function name and qualifier are required to get the resource.
        function_name(str, Optional): The name of the Lambda function, version, or alias. If not supplied, the function name will be extracted fom the resource_id.
        qualifier(str, Optional): A version number or alias name.  If not supplied, the qualifier will be extracted fom the resource_id.

    Returns:
        .. code-block:: python

           {"result": True|False, "comment": A message List, "ret": None|Dict}

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec aws.lambda_aws.function_provisioned_concurrency_config.get resource_id="resource_id" name="name"


        Calling this exec module function from the cli with function_name and qualifier.

        .. code-block:: bash

            idem exec aws.lambda_aws.function_provisioned_concurrency_config.get function_name="function_name" qualifier="qualifier" name="name"


        Calling this exec module function from within a state module in pure python

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await aws.lambda_aws.function_provisioned_concurrency_config.get(ctx, resource_id=resource_id, name=name, **kwargs)
    """
    result = dict(comment=[], ret=None, result=True)

    if resource_id:
        function_name = hub.tool.aws.arn_utils.get_resource_name(resource_id)
        qualifier = hub.tool.aws.arn_utils.get_qualifier(resource_id)

    before = await hub.exec.boto3.client["lambda"].get_provisioned_concurrency_config(
        ctx, FunctionName=function_name, Qualifier=qualifier
    )

    if not before["result"]:
        if "ProvisionedConcurrencyConfigNotFoundException" in str(before["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.lambda.function_provisioned_concurrency_config",
                    name=name,
                )
            )
            result["comment"] += list(before["comment"])
            return result
        result["result"] = False
        result["comment"] = list(before["comment"])
        return result

    # get_provisioned_concurrency_config method does not return function ARN. Hence build function ARN which will be set as resource_id
    account_details = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
    account_id = account_details["ret"]["Account"]
    region_name = ctx["acct"]["region_name"]
    function_arn = hub.tool.aws.arn_utils.build(
        service="lambda",
        region=region_name,
        account_id=account_id,
        resource="function:" + function_name + ":" + qualifier,
    )

    result[
        "ret"
    ] = hub.tool.aws.lambda_aws.function_provisioned_concurrency_config.convert_raw_function_concurrency_config_to_present(
        before["ret"],
        function_name=function_name,
        idem_resource_name=name,
        function_arn=function_arn,
    )

    return result


async def list_(hub, ctx, function_name: str):
    """Retrieves a list of provisioned concurrency configurations for a function.

    Args:
        function_name(str):
            The name of the Lambda function.
                Name formats
                    * Function name - my-function.
                    * Function ARN - ``arn:aws:lambda:us-west-2:123456789012:function:my-function``.
                    * Partial ARN - ``123456789012:function:my-function``.

    Returns:
        .. code-block:: python

           {"result": True|False, "comment": A message List, "ret": None|Dict}

    Examples:
        Calling this exec module function from the cli with function_name

        .. code-block:: bash

            idem exec aws.lambda_aws.function_provisioned_concurrency_config.list function_name="function_name"


        Calling this exec module function from within a state module in pure python

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await aws.lambda_aws.function_provisioned_concurrency_config.list(ctx, function_name=function_name)
    """
    result = dict(comment=[], ret=[], result=True)

    list_provisioned_concurrency_configs_ret = await hub.exec.boto3.client[
        "lambda"
    ].list_provisioned_concurrency_configs(ctx, FunctionName=function_name)
    if not list_provisioned_concurrency_configs_ret["result"]:
        result["comment"] += list(list_provisioned_concurrency_configs_ret["comment"])
        result["result"] = False
        return result

    for (
        function_provisioned_concurrency_config
    ) in list_provisioned_concurrency_configs_ret["ret"][
        "ProvisionedConcurrencyConfigs"
    ]:
        result["ret"].append(
            hub.tool.aws.lambda_aws.function_provisioned_concurrency_config.convert_raw_function_concurrency_config_to_present(
                function_provisioned_concurrency_config,
                function_name=function_name,
            )
        )
    return result



"""State module for managing Amazon KMS Alias."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": [
            "aws.kms.key.absent",
        ],
    },
    "present": {
        "require": [
            "aws.kms.key.present",
        ],
    },
}


async def present(
    hub, ctx, name: str, target_key_id: str, resource_id: str = None
) -> Dict[str, Any]:
    """Creates a friendly name for a KMS key.

    You can associate the alias with any customer managed key in the same Amazon Web Services Region.
    Each alias is associated with only one KMS key at a time, but a KMS key can have multiple aliases.
    A valid KMS key is required. You can't create an alias without a KMS key.

    The alias must be unique in the account and Region, but you can have aliases with the same name in different Regions

    Args:
        name(str):
            The name of the alias. This value must begin with ``alias/`` followed by a name, such as ``alias/ExampleAlias``.
            The AliasName value must be string of 1-256 characters. It can contain only alphanumeric characters, forward slashes (``/``),
            underscores (``_``), and dashes (``-``). The alias name cannot begin with ``alias/aws/``. The ``alias/aws/`` prefix is reserved
            for Amazon Web Services managed keys.
        target_key_id(str):
            Associates the alias with the specified customer managed key . The KMS key must be in the same Amazon Web Services Region..
        resource_id(str, Optional):
            The name of the alias in Amazon Web Services.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_kms_alias]:
          aws.kms.alias.present:
            - name: 'string'
            - target_key_id: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_kms_alias:
              aws.kms.alias.present:
                  - name: alias/my-kms-key
                  - target_key_id: 1234abcd-12ab-34cd-56ef-1234567890ab
    """
    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    # check if alias name already exists
    before = await hub.tool.aws.kms.alias.get_alias_by_name(
        ctx, resource_id if resource_id else name
    )
    is_alias_applied = False

    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result
    elif before["ret"]:
        result[
            "old_state"
        ] = hub.tool.aws.kms.conversion_utils.convert_raw_key_alias_to_present(
            raw_resource=before["ret"]
        )
        plan_state = copy.deepcopy(result["old_state"])
        # alias already exists with that name. Update alias if target key is different
        if result["old_state"].get("target_key_id") != target_key_id:
            try:
                if not ctx.get("test", False):
                    update_ret = await hub.exec.boto3.client.kms.update_alias(
                        ctx, AliasName=name, TargetKeyId=target_key_id
                    )
                    result["result"] = update_ret["result"]
                    if not result["result"]:
                        result["comment"] = update_ret["comment"]
                        return result
                    result[
                        "comment"
                    ] = f"Updated alias of KMS key {target_key_id} to {name}"
                    is_alias_applied = True
                else:
                    result["comment"] = f"Would update aws.kms.alias {name}"
                    plan_state["target_key_id"] = target_key_id

            except hub.tool.boto3.exception.ClientError as e:
                result["comment"] = f"{e.__class__.__name__}: {e}"
                result["result"] = False
                return result
        else:
            result[
                "comment"
            ] = f"KMS Alias '{name}' is already associated with the KMS target key '{target_key_id}'"
    else:
        # No alias exists with that name. create new alias

        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "target_key_id": target_key_id,
                },
            )
            result["comment"] = f"Would create aws.kms.alias {name}"
            return result

        try:
            ret = await hub.exec.boto3.client.kms.create_alias(
                ctx, AliasName=name, TargetKeyId=target_key_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result[
                "comment"
            ] = f"Created aws.kms.alias {name} for KMS key aws.kms.key {target_key_id}"
            is_alias_applied = True
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False
            return result
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        # get new state only if we create or modify alias
        elif is_alias_applied:
            after = await hub.tool.aws.kms.alias.get_alias_by_name(ctx, name)
            if after["result"] and after.get("ret"):
                result[
                    "new_state"
                ] = hub.tool.aws.kms.conversion_utils.convert_raw_key_alias_to_present(
                    raw_resource=after["ret"]
                )
            else:
                result["comment"] = after["comment"]
                result["result"] = False
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes an AWS KMS alias.

    Because an alias is not a property of a KMS key, you can delete and change the aliases of a KMS key without
    affecting the KMS key.

    Args:
        name(str):
            The name of the alias.
        resource_id(str, Optional):
            The name of the alias in Amazon Web Services.

    Returns:
        Dict[str, Any]

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_kms_alias]:
          aws.kms.alias.absent:
            - name: 'string'
            - resource_id: 'string'

    Examples:
        .. code-block:: sls

            idem_test_aws_kms_alias:
              aws.kms.alias.absent:
                - name: alias/my-kms-key
                - resource_id: alias/my-kms-key
    """
    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)

    before = await hub.tool.aws.kms.alias.get_alias_by_name(
        ctx, (resource_id if resource_id else name)
    )
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    if not before["ret"]:
        result["comment"] = f"aws.kms.policy '{name}' already absent"
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.kms.conversion_utils.convert_raw_key_alias_to_present(
            raw_resource=before["ret"]
        )
        result["comment"] = f"Would delete aws.kms.alias '{name}'"
    else:
        result[
            "old_state"
        ] = hub.tool.aws.kms.conversion_utils.convert_raw_key_alias_to_present(
            raw_resource=before["ret"]
        )
        try:
            ret = await hub.exec.boto3.client.kms.delete_alias(
                ctx, AliasName=(resource_id if resource_id else name)
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = f"aws.kms.alias '{name}' is deleted"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes AWS KMS alias in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.kms.alias
    """
    result = {}
    ret = await hub.exec.boto3.client.kms.list_aliases(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.kms.alias {ret['comment']}")
        return {}

    for key in ret["ret"]["Aliases"]:
        # Remove AWS reserved aliases as we cannot modify them using present/absent
        if key.get("AliasName") is not None and "alias/aws/" in key.get("AliasName"):
            continue
        # Get alias details to match the 'present' function parameters
        translated_resource = (
            hub.tool.aws.kms.conversion_utils.convert_raw_key_alias_to_present(
                raw_resource=key
            )
        )
        result[translated_resource["resource_id"]] = {
            "aws.kms.alias.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result



from typing import Any
from typing import Dict
from typing import List


async def update(
    hub,
    ctx,
    resource_id: str,
    before: Dict[str, Any],
    users: List,
):
    """Updates association of users in IAM Group.

    Args:
        hub: required for functions in hub.
        ctx: context.
        resource_id: AWS IAM Group name.
        before(Dict): Existing resource parameters in Amazon Web Services.
        users(List): List of users

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}

    """
    result = dict(comment=(), result=True, ret=None)
    if set(before.get("users")) != set(users):
        users_to_add = list(set(users).difference(before.get("users")))
        users_to_remove = list(set(before.get("users")).difference(users))

        if users_to_remove:
            for user in users_to_remove:
                if not ctx.get("test", False):
                    remove_usr_ret = (
                        await hub.exec.boto3.client.iam.remove_user_from_group(
                            ctx=ctx, GroupName=resource_id, UserName=user
                        )
                    )
                    if not remove_usr_ret["result"]:
                        result["comment"] = remove_usr_ret["comment"]
                        result["result"] = False
                        return result

        if users_to_add:
            for user in users_to_add:
                if not ctx.get("test", False):
                    add_usr_ret = await hub.exec.boto3.client.iam.add_user_to_group(
                        ctx, GroupName=resource_id, UserName=user
                    )
                    if not add_usr_ret["result"]:
                        result["comment"] = add_usr_ret["comment"]
                        result["result"] = False
                        return result

        result["ret"] = {}
        result["ret"]["users"] = users
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.iam.group_membership", name=resource_id
            )
        else:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.iam.group_membership", name=resource_id
            )

    return result



from typing import List


async def apply(
    hub, ctx, resource, *, old_value, new_value, comments: List[str]
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in it's "present" state

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    if new_value is None:
        ret = await hub.exec.boto3.client.ec2.reset_instance_attribute(
            ctx, InstanceId=resource.id, Attribute="instanceInitiatedShutdownBehavior"
        )
        if ret.comment:
            comments.append(ret.comment)
        return ret.result
    else:
        ret = await hub.exec.boto3.client.ec2.modfiy_instance_attribute(
            ctx,
            InstanceId=resource.id,
            Attribute="instanceInitiatedShutdownBehavior",
            InstanceInitiatedShutdownBehavior={"Value": new_value},
        )
        if ret.comment:
            comments.append(ret.comment)
        return ret.result



def validation_rec(hub, validation_rec_fqdns, certificate):
    """
    Implement additional sanity checks for all FQDNs
    Args:
        hub:
        validation_rec_fqdns(List): List of FQDNs that implement the validation.
        certificate(Dict): Certificate obtained from describe.

    Returns:
        {"result": True|False, "comment": "A tuple"}

    """
    resource_id = certificate.get("resource_id")
    expected_fqdns = {}
    ret = dict(result=True, comment=())
    if not certificate.get("DomainValidationOptions"):
        ret["comment"] = (f"domain_validation_options empty for {resource_id}",)
        ret["result"] = False
        return ret
    for validation_option in certificate.get("DomainValidationOptions"):
        if (
            validation_option.get("ValidationMethod")
            and validation_option.get("ValidationMethod")
            != hub.states.aws.acm.certificate_validation.DNS
        ):
            hub.log.debug("validation_record_fqdns is only valid for DNS validation")
            ret["comment"] = ret["comment"] + (
                "validation_record_fqdns is only valid for DNS validation",
            )
            ret["result"] = False
            return ret
        if validation_option.get("ResourceRecord") and validation_option.get(
            "ResourceRecord"
        ).get("Name"):
            new_expected_fqdn = (
                validation_option.get("ResourceRecord").get("Name").rstrip(".")
            )
            expected_fqdns[new_expected_fqdn] = validation_option
    for v in validation_rec_fqdns:
        expected_fqdns.pop(v.rstrip("."))
    if expected_fqdns:
        for expected_fqdn in expected_fqdns:
            hub.log.debug(
                f"missing DNS validation record {expected_fqdns.get(expected_fqdn).get('DomainName')} : {expected_fqdn}"
            )
            ret["comment"] = ret["comment"] + (
                f"missing DNS validation record {expected_fqdns.get(expected_fqdn).get('DomainName')} : {expected_fqdn}",
            )
        ret["result"] = False
    return ret


def extract_validation_method(hub, certificate):
    validation_method = None
    if certificate.get("DomainValidationOptions"):
        for validation_option in certificate.get("DomainValidationOptions"):
            if "ValidationMethod" in validation_option:
                validation_method = validation_option.get("ValidationMethod")
                return validation_method
    return validation_method
